# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinn_utilities.typing.coords import XYP
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, AbstractSCPResponse)
from spinnman.messages.scp.enums import SCPCommand, SCPResult
from spinnman.messages.sdp import SDPFlag, SDPHeader
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException
from spinnman.constants import address_length_dtype


class Response(AbstractSCPResponse):
    """
    An SCP response to a request to read a region of memory on a chip.
    """
    __slots__ = (
        "_data",
        "_length",
        "_offset",
        "__op",
        "__cmd")

    def __init__(self, operation: str, command: str) -> None:
        super().__init__()
        self._data = b''
        self._length = 0
        self._offset = 0
        self.__op = operation
        self.__cmd = command

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data: bytes, offset: int):
        assert self._scp_response_header is not None
        if self._scp_response_header.result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                self.__op, self.__cmd, self._scp_response_header.result)
        self._data = data
        self._offset = offset
        self._length = len(data) - offset

    @property
    def data(self) -> bytes:
        """
        The data read.

        .. note::
            The data starts at offset.

        :rtype: bytearray
        """
        return self._data

    @property
    def offset(self) -> int:
        """
        The offset where the valid data starts.

        :rtype: int
        """
        return self._offset

    @property
    def length(self) -> int:
        """
        The length of the valid data.

        :rtype: int
        """
        return self._length


class ReadMemory(AbstractSCPRequest[Response]):
    """
    An SCP request to read a region of memory on a chip.
    """
    __slots__ = ()

    def __init__(self, coordinates: XYP, base_address: int, size: int):
        """
        :param tuple coordinates:
            The X,Y,P coordinates of the chip to read from;
            X and Y between 0 and 255, P between 0 and 17
        :param int base_address:
            The positive base address to start the read from
        :param int size: The number of bytes to read, between 1 and 256
        :raise SpinnmanInvalidParameterException:
            * If the chip coordinates are out of range
            * If the base address is not a positive number
            * If the size is out of range
        """
        x, y, cpu = coordinates
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=cpu, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_READ),
            argument_1=base_address, argument_2=size,
            argument_3=address_length_dtype[
                (base_address % 4, size % 4)].value)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> Response:
        return Response("read memory", "CMD_READ")



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import IntEnum


class SpallocState(IntEnum):
    """
    The possible states of a Spalloc Job.

    Jobs start as `QUEUED`, then move to `POWER` once they've been allocated.
    Once the job's boards have been switched on and the hardware stabilised,
    the job moves to state `READY`. (It goes back to `POWER` if you tell it to
    switch off or on; this is not recommended.) Finally, it goes to `DESTROYED`
    once the job is completed in any way.

    The `UNKNOWN` state is used when something odd is going on. It should
    normally be ignored.
    """
    #: The job is in an unknown state.
    UNKNOWN = 0
    #: The job is queued waiting for allocation.
    QUEUED = 1
    #: The job is queued waiting for boards to power on or off.
    POWER = 2
    #: The job is ready for user code to run on it.
    READY = 3
    #: The job has been destroyed.
    DESTROYED = 4



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class MachineDimensions(object):
    """
    Represents the size of a machine in chips.
    """
    __slots__ = [
        "_height",
        "_width"]

    def __init__(self, width: int, height: int):
        """
        :param int width: The width of the machine in chips
        :param int height: The height of the machine in chips
        """
        self._width = width
        self._height = height

    @property
    def width(self) -> int:
        """
        The width of the machine in chips.

        :rtype: int
        """
        return self._width

    @property
    def height(self) -> int:
        """
        The height of the machine in chips.

        :rtype: int
        """
        return self._height



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional
from spinn_utilities.overrides import overrides
from spinnman.messages.scp.abstract_messages import AbstractSCPResponse
from spinnman.messages.scp.enums import SCPResult
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException
from spinnman.model import ChipSummaryInfo


class GetChipInfoResponse(AbstractSCPResponse):
    """
    An SCP response to a request for the version of software running.
    """
    __slots__ = "_chip_info",

    def __init__(self) -> None:
        super().__init__()
        self._chip_info: Optional[ChipSummaryInfo] = None

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data: bytes, offset: int):
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                "Version", "CMD_CHIP_INFO", result.name)
        self._chip_info = ChipSummaryInfo(
            data, offset, self.sdp_header.source_chip_x,
            self.sdp_header.source_chip_y,)

    @property
    def chip_info(self) -> ChipSummaryInfo:
        """
        The chip information received.

        :rtype: ChipSummaryInfo
        """
        assert self._chip_info is not None
        return self._chip_info



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import struct
from spinn_utilities.overrides import overrides
from spinnman.constants import EIEIO_COMMAND_IDS
from .eieio_command_message import EIEIOCommandMessage
from .eieio_command_header import EIEIOCommandHeader

_PATTERN_B = struct.Struct("<B")


class HostDataReadAck(EIEIOCommandMessage):
    """
    Packet sent by the host computer to the SpiNNaker system in the
    context of the buffering output technique to signal that the host has
    received a request to read data.
    """
    __slots__ = "_sequence_no",

    def __init__(self, sequence_no: int):
        """

        :param int sequence_no:
        """
        super().__init__(
            EIEIOCommandHeader(EIEIO_COMMAND_IDS.HOST_DATA_READ_ACK))
        self._sequence_no = sequence_no

    @property
    def sequence_no(self) -> int:
        """
        Gets the sequence_no passed into the init.

        :rtype: int
        """
        return self._sequence_no

    @staticmethod
    @overrides(EIEIOCommandMessage.from_bytestring)
    def from_bytestring(command_header: EIEIOCommandHeader, data: bytes,
                        offset: int) -> "HostDataReadAck":
        sequence_no = _PATTERN_B.unpack_from(data, offset)[0]

        return HostDataReadAck(sequence_no)

    @property
    @overrides(EIEIOCommandMessage.bytestring)
    def bytestring(self) -> bytes:
        byte_string = super().bytestring
        byte_string += _PATTERN_B.pack(self.sequence_no)
        return byte_string



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time


class TokenBucket(object):
    """
    An implementation of the token bucket algorithm. Usage::

        >>> bucket = TokenBucket(80, 0.5)
        >>> print(bucket.consume(10))
        True

    Not thread safe.
    """
    __slots__ = ('_capacity', '_tokens', '_fill_rate', '_timestamp')

    def __init__(self, tokens, fill_rate):
        """
        :param int tokens: the total tokens in the bucket
        :param float fill_rate:
            the rate in tokens/second that the bucket will be refilled.
        """
        self._capacity = float(tokens)
        self._tokens = float(tokens)
        self._fill_rate = float(fill_rate)
        self._timestamp = time.time()

    def consume(self, tokens, block=True):
        """
        Consume tokens from the bucket. Returns True if there were
        sufficient tokens.

        If there are not enough tokens and block is True, sleeps until the
        bucket is replenished enough to satisfy the deficiency.

        If there are not enough tokens and block is False, returns False.

        It is an error to consume more tokens than the bucket capacity.

        :param int tokens:
        :param bool block:
        :rtype: bool
        """
        while block and tokens > self.tokens:
            deficit = tokens - self._tokens
            delay = deficit / self._fill_rate
            time.sleep(delay)

        if tokens <= self.tokens:
            self._tokens -= tokens
            return True
        return False

    @property
    def tokens(self):
        """
        The number of tokens currently in the bucket.

        :rtype: int
        """
        if self._tokens < self._capacity:
            now = time.time()
            delta = self._fill_rate * (now - self._timestamp)
            self._tokens = min(self._capacity, self._tokens + delta)
            self._timestamp = now
        return self._tokens



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import logging
from threading import Thread
from typing import Callable, Generic, List, TypeVar
from concurrent.futures import ThreadPoolExecutor, Future
from spinn_utilities.log import FormatAdapter
from spinnman.exceptions import SpinnmanEOFException
from spinnman.connections.abstract_classes import Listenable

#: :meta private:
T = TypeVar("T")
logger = FormatAdapter(logging.getLogger(__name__))
_POOL_SIZE = 4
_TIMEOUT = 1


class ConnectionListener(Thread, Generic[T]):
    """
    Thread that listens to a connection and calls callbacks with new
    messages when they arrive.
    """
    __slots__ = (
        "__callback_pool",
        "__callbacks",
        "__connection",
        "__done",
        "__timeout")

    def __init__(self, connection: Listenable[T],
                 n_processes: int = _POOL_SIZE, timeout: float = _TIMEOUT):
        """
        :param Listenable connection: A connection to listen to
        :param int n_processes:
            The number of threads to use when calling callbacks
        :param float timeout:
            How long to wait for messages before checking to see if the
            connection is to be terminated.
        """
        super().__init__(
            name=f"Connection listener for connection {connection}")
        self.daemon = True
        self.__connection = connection
        self.__timeout = timeout
        self.__callback_pool = ThreadPoolExecutor(max_workers=n_processes)
        self.__done = False
        self.__callbacks: List[Callable[[T], None]] = []

    def __run_step(self, handler: Callable[[], T]):
        """
        :param ~collections.abc.Callable handler:
        """
        if self.__connection.is_ready_to_receive(timeout=self.__timeout):
            message = handler()
            for callback in self.__callbacks:
                future = self.__callback_pool.submit(
                    callback, message)
                future.add_done_callback(self.__done_callback)

    def __done_callback(self, future: Future[None]):
        """
        :param ~concurrent.futures.Future future:
        """
        try:
            future.result()
        except Exception:  # pylint: disable=broad-except
            logger.exception("problem in listener call")

    def run(self) -> None:
        """
        Implements the listening thread.
        """
        with self.__callback_pool:
            handler = self.__connection.get_receive_method()
            while not self.__done:
                try:
                    self.__run_step(handler)
                except SpinnmanEOFException:
                    self.__done = True
                except Exception:  # pylint: disable=broad-except
                    if not self.__done:
                        logger.warning("problem when dispatching message",
                                       exc_info=True)

    def add_callback(self, callback: Callable[[T], None]):
        """
        Add a callback to be called when a message is received.

        :param ~collections.abc.Callable callback:
            A callable which takes a single parameter, which is the message
            received; the result of the callback will be ignored.
        """
        self.__callbacks.append(callback)

    def close(self) -> None:
        """
        Closes the listener.

        .. note::
            This does not close the provider of the messages; this instead
            marks the listener as closed.  The listener will not truly stop
            until the get message call returns.
        """
        self.__done = True
        self.join()



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Mapping
from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.sdp import SDPFlag, SDPHeader
from spinnman.messages.scp.impl.check_ok_response import CheckOKResponse


class SetLED(AbstractSCPRequest[CheckOKResponse]):
    """
    A request to change the state of a chip's LED.

    This class is currently deprecated and untested as there is no
    known use except for Transceiver.set_led which is itself deprecated.
    """
    __slots__ = ()

    def __init__(
            self, x: int, y: int, cpu: int, led_states: Mapping[int, int]):
        """
        :param int x: The x-coordinate of the chip, between 0 and 255
        :param int y: The y-coordinate of the chip, between 0 and 255
        :param int cpu:
            The CPU-number to use to set the LED. Normally should be 0
        :param dict(int,int) led_states:
            A dictionary mapping LED index to state with
            0 being off, 1 on and 2 inverted.
        """
        encoded_led_states = 0
        for led, state in led_states.items():
            encoded_led_states |= {0: 2, 1: 3, 2: 1}[state] << (2 * led)

        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=cpu, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_LED),
            argument_1=encoded_led_states)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> CheckOKResponse:
        return CheckOKResponse("Set chip LED", "CMD_LED")



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from spinn_utilities.overrides import overrides
from spinnman.messages.scp.abstract_messages import AbstractSCPResponse
from spinnman.messages.scp.enums import SCPResult
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException

_ONE_WORD = struct.Struct("<I")


class CountStateResponse(AbstractSCPResponse):
    """
    An SCP response to a request for the number of cores in a given state.
    """
    __slots__ = "_count",

    def __init__(self):
        super().__init__()
        self._count = None

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data: bytes, offset: int):
        result = self.scp_response_header.result
        # We can accept a no-reply response here; that could just mean
        # that the count wasn't complete (but might be enough anyway)
        if result != SCPResult.RC_OK and result != SCPResult.RC_P2P_NOREPLY:
            raise SpinnmanUnexpectedResponseCodeException(
                "CountState", "CMD_COUNT", result.name)
        self._count = _ONE_WORD.unpack_from(data, offset)[0]

    @property
    def count(self) -> int:
        """
        The count of the number of cores with the requested state.

        :rtype: int
        """
        return self._count



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, List
from spinn_utilities.overrides import overrides
from spinn_utilities.typing.coords import XY
from spinnman.data import SpiNNManDataView
from spinnman.connections.udp_packet_connections import SCAMPConnection
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)


class MostDirectConnectionSelector(ConnectionSelector):
    """
    A selector that goes for the most direct connection for the message.
    """
    __slots__ = (
        "_connections",
        "_lead_connection")

    def __init__(self, connections: List[SCAMPConnection]):
        """
        :param list(SCAMPConnection) connections:
            The connections to be used
        """
        self._connections: Dict[XY, SCAMPConnection] = dict()
        lead_connection = None
        for conn in connections:
            if conn.chip_x == 0 and conn.chip_y == 0:
                lead_connection = conn
            self._connections[conn.chip_x, conn.chip_y] = conn
        if lead_connection is None:
            lead_connection = next(iter(connections))
        self._lead_connection = lead_connection

    @overrides(ConnectionSelector.get_next_connection)
    def get_next_connection(
            self, message: AbstractSCPRequest) -> SCAMPConnection:
        key = (message.sdp_header.destination_chip_x,
               message.sdp_header.destination_chip_y)
        if key in self._connections:
            return self._connections[key]

        if not SpiNNManDataView.has_machine() or len(self._connections) == 1:
            return self._lead_connection

        x, y = key
        key = SpiNNManDataView.get_nearest_ethernet(x, y)

        return self._connections.get(key, self._lead_connection)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=too-many-arguments

from typing import (
    BinaryIO, Collection, Dict, FrozenSet, Iterable,
    List, Optional, Set, Tuple, Union)
from spinn_utilities.abstract_base import abstractmethod
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.typing.coords import XY
from spinn_machine import (
    CoreSubsets, Machine, MulticastRoutingEntry, RoutingEntry)
from spinn_machine.tags import AbstractTag, IPTag, ReverseIPTag
from spinnman.connections.abstract_classes import Connection
from spinnman.connections.udp_packet_connections import (
    SCAMPConnection, SDPConnection)
from spinnman.messages.scp.enums import Signal
from spinnman.messages.sdp import SDPMessage
from spinnman.model import (
    CPUInfos, DiagnosticFilter, IOBuffer, RouterDiagnostics,
    VersionInfo)
from spinnman.model.enums import CPUState, UserRegister
from spinnman.processes import MostDirectConnectionSelector


class Transceiver(object):
    """
    An encapsulation of various communications with the SpiNNaker board.

    The methods of this class are designed to be thread-safe (provided they do
    not access a BMP, as access to those is never thread-safe);
    thus you can make multiple calls to the same (or different) methods
    from multiple threads and expect each call to work as if it had been
    called sequentially, although the order of returns is not guaranteed.

    .. note::
        With multiple connections to the board, using multiple threads in this
        way may result in an increase in the overall speed of operation, since
        the multiple calls may be made separately over the set of given
        connections.
    """
    __slots__ = ()

    @abstractmethod
    def send_sdp_message(self, message: SDPMessage,
                         connection: Optional[SDPConnection] = None):
        """
        Sends an SDP message using one of the connections.

        :param SDPMessage message: The message to send
        :param SDPConnection connection: An optional connection to use
        """
        # https://github.com/SpiNNakerManchester/SpiNNMan/issues/369
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def discover_scamp_connections(self) -> None:
        """
        Find connections to the board and store these for future use.

        .. note::
            An exception will be thrown if no initial connections can be
            found to the board.

        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        # Used directly after Transceiver init
        # Not called if add_scamp_connections is called
        # Not called by SpallocJobController
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def add_scamp_connections(self, connections: Dict[XY, str]):
        """
        Check connections to the board and store these for future use.

        .. note::
            An exception will be thrown if no initial connections can be
            found to the board.

        :param dict((int,int),str) connections:
            Dict of (`x`,`y`) to IP address
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        # Use on a spalloc created Transceiver
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def get_connections(self) -> Set[Connection]:
        """
        Get the currently known connections to the board, made up of those
        passed in to the transceiver and those that are discovered during
        calls to discover_connections.  No further discovery is done here.

        :return: An iterable of connections known to the transceiver
        :rtype: set(Connection)
        """
        # used in unittest only
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def get_machine_details(self) -> Machine:
        """
        Get the details of the machine made up of chips on a board and how
        they are connected to each other.

        :return: A machine description
        :rtype: ~spinn_machine.Machine
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        # used by machine_generator
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def get_cpu_infos(
            self, core_subsets: Optional[CoreSubsets] = None,
            states: Union[CPUState, Iterable[CPUState], None] = None,
            include: bool = True) -> CPUInfos:
        """
        Get information about the processors on the board.

        :param ~spinn_machine.CoreSubsets core_subsets:
            A set of chips and cores from which to get the
            information. If not specified, the information from all of the
            cores on all of the chips on the board are obtained.
        :param states: The state or states to filter on (if any)
        :type states: None, CPUState or iterable(CPUState)
        :param bool include:
            If `True` includes only infos in the requested state(s).
            If `False` includes only infos *not* in the requested state(s).
            Ignored if states is `None`.
        :return: The CPU information for the selected cores and States, or
            all cores/states  if core_subsets/states is not specified
        :rtype: ~spinnman.model.CPUInfos
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If chip_and_cores contains invalid items
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        # used by
        # application_finisher
        # chip_provenance_updater
        # emergency_recover_state_from_failure
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def get_clock_drift(self, x: int, y: int) -> float:
        """
        Get the clock drift.

        :param int x: The x-coordinate of the chip to get drift for
        :param int y: The y-coordinate of the chip to get drift for
        """
        # used by drift_report
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def read_user(self, x: int, y: int, p: int, user: UserRegister):
        """
        Get the contents of the this user register for the given processor.

        .. note::
            Conventionally, user_0 usually holds the address of the table of
            memory regions.

        :param int x: X coordinate of the chip
        :param int y: Y coordinate of the chip
        :param int p: Virtual processor identifier on the chip
        :param int user: The user number to read data for
        :rtype: int
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            If x, y, p does not identify a valid processor
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        # used by pair_compression and others during _check_for_success
        # used by memory_map_on_host_chip_report._get_region_table_addr
        #   https://github.com/SpiNNakerManchester/SpiNNFrontEndCommon/pull/1104
        # used by DataSpeedUpPacketGatherMachineVertex
        #    and ExtraMonitorSupportMachineVertex
        #    to .update_transaction_id_from_machine
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def add_cpu_information_from_core(
            self, cpu_infos: CPUInfos, x: int, y: int, p: int,
            states: Iterable[CPUState]):
        """
        Adds information about a specific processor on the board to the info

        :param CPUInfos cpu_infos: Info to add data for this core to
        :param int x: The x-coordinate of the chip containing the processor
        :param int y: The y-coordinate of the chip containing the processor
        :param int p: The ID of the processor to get the information about
        :param states:
            If provided will only add the info if in one of the states
        :type states: list(CPUState)
        :return: The CPU information for the selected core
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If x, y, p is not a valid processor
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        # used by emergency_recover_state_from_failure
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def get_region_base_address(self, x: int, y: int, p: int):
        """
        Gets the base address of the Region Table

        :param int x: The x-coordinate of the chip containing the processor
        :param int y: The y-coordinate of the chip containing the processor
        :param int p: The ID of the processor to get the address
        :return: The address of the Region table for the selected core
        :rtype: int
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If x, y, p is not a valid processor
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def get_iobuf(self, core_subsets: Optional[CoreSubsets] = None
                  ) -> Iterable[IOBuffer]:
        """
        Get the contents of the IOBUF buffer for a number of processors.

        :param ~spinn_machine.CoreSubsets core_subsets:
            A set of chips and cores from which to get the buffers. If not
            specified, the buffers from all of the cores on all of the chips
            on the board are obtained.
        :return: An iterable of the buffers, which may not be in the order
            of core_subsets
        :rtype: iterable(IOBuffer)
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If chip_and_cores contains invalid items
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        # Used by IOBufExtractor
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def get_core_state_count(
            self, app_id: int, state: CPUState,
            xys: Optional[Iterable[Tuple[int, int]]] = None) -> int:
        """
        Get a count of the number of cores which have a given state.

        :param int app_id:
            The ID of the application from which to get the count.
        :param CPUState state: The state count to get
        :param list(int,int) xys: The chips to query, or None for all
        :return: A count of the cores with the given status
        :rtype: int
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If state is not a valid status
            * If app_id is not a valid application ID
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def execute_flood(
            self, core_subsets: CoreSubsets,
            executable: Union[BinaryIO, bytes, str], app_id: int, *,
            n_bytes: Optional[int] = None, wait: bool = False):
        """
        Start an executable running on multiple places on the board.  This
        will be optimised based on the selected cores, but it may still
        require a number of communications with the board to execute.

        :param ~spinn_machine.CoreSubsets core_subsets:
            Which cores on which chips to start the executable
        :param executable:
            The data that is to be executed. Should be one of the following:
            * An instance of RawIOBase
            * A bytearray
            * A filename of an executable
        :type executable:
            ~io.RawIOBase or bytes or bytearray or str
        :param int app_id:
            The ID of the application with which to associate the executable
        :param int n_bytes:
            The size of the executable data in bytes. If not specified:

            * If `executable` is an RawIOBase, an error is raised
            * If `executable` is a bytearray, the length of the bytearray will
              be used
            * If `executable` is a str, the length of the file will be used
        :param bool wait:
            True if the processors should enter a "wait" state on loading
        :raise SpinnmanIOException:
            * If there is an error communicating with the board
            * If there is an error reading the executable
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If one of the specified cores is not valid
            * If `app_id` is an invalid application ID
            * If a packet is received that has invalid parameters
            * If `executable` is an RawIOBase but `n_bytes` is not specified
            * If `executable` is an int and `n_bytes` is more than 4
            * If `n_bytes` is less than 0
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        # Used by load_app_images, load_sys_images and
        # run_system_application._load_application
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def read_fpga_register(
            self, fpga_num: int, register: int, board: int = 0) -> int:
        """
        Read a register on a FPGA of a board. The meaning of the
        register's contents will depend on the FPGA's configuration.

        :param int fpga_num: FPGA number (0, 1 or 2) to communicate with.
        :param int register:
            Register address to read to (will be rounded down to
            the nearest 32-bit word boundary).
        :param int board: which board to request the FPGA register from
        :return: the register data
        :rtype: int
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def write_fpga_register(
            self, fpga_num: int, register: int, value: int, board: int = 0):
        """
        Write a register on a FPGA of a board. The meaning of setting the
        register's contents will depend on the FPGA's configuration.

        :param int fpga_num: FPGA number (0, 1 or 2) to communicate with.
        :param int register:
            Register address to read to (will be rounded down to
            the nearest 32-bit word boundary).
        :param int value: the value to write into the FPGA register
        :param int board: which board to write the FPGA register to
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def read_bmp_version(self, board: int) -> VersionInfo:
        """
        Read the BMP version.

        :param int board: which board to request the data from
        :return: the version_info from the BMP
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def write_memory(
            self, x: int, y: int, base_address: int,
            data: Union[BinaryIO, bytes, int, str], *,
            n_bytes: Optional[int] = None, offset: int = 0, cpu: int = 0,
            get_sum: bool = False) -> Tuple[int, int]:
        """
        Write to the SDRAM on the board.

        :param int x:
            The x-coordinate of the chip where the memory is to be written to
        :param int y:
            The y-coordinate of the chip where the memory is to be written to
        :param int base_address:
            The address in SDRAM where the region of memory is to be written
        :param data: The data to write.  Should be one of the following:

            * An instance of RawIOBase
            * A bytearray/bytes
            * A single integer - will be written in little-endian byte order
            * A string - the filename of a data file
        :type data:
            ~io.RawIOBase or bytes or bytearray or int or str
        :param int n_bytes:
            The amount of data to be written in bytes.  If not specified:

            * If `data` is an RawIOBase, an error is raised
            * If `data` is a bytearray, the length of the bytearray will be
              used
            * If `data` is an int, 4 will be used
            * If `data` is a str, the length of the file will be used
        :param int offset: The offset from which the valid data begins
        :param int cpu: The optional CPU to write to
        :param bool get_sum: whether to return a checksum or 0
        :return: The number of bytes written, the checksum (0 if get_sum=False)
        :rtype: int, int
        :raise SpinnmanIOException:
            * If there is an error communicating with the board
            * If there is an error reading the data
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If `x, y` does not lead to a valid chip
            * If a packet is received that has invalid parameters
            * If `base_address` is not a positive integer
            * If `data` is an RawIOBase but `n_bytes` is not specified
            * If `data` is an int and `n_bytes` is more than 4
            * If `n_bytes` is less than 0
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def write_user(
            self, x: int, y: int, p: int, user: UserRegister, value: int):
        """
        Write to the user *N* "register" for the given processor.

        .. note::
            Conventionally, user_0 usually holds the address of the table of
            memory regions.

        :param int x: X coordinate of the chip
        :param int y: Y coordinate of the chip
        :param int p: Virtual processor identifier on the chip
        :param int user: The user "register" number of write data for
        :param int value: The value to write
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            If x, y, p does not identify a valid processor
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def read_memory(
            self, x: int, y: int, base_address: int, length: int,
            cpu: int = 0) -> bytearray:
        """
        Read some areas of memory (usually SDRAM) from the board.

        :param int x:
            The x-coordinate of the chip where the memory is to be read from
        :param int y:
            The y-coordinate of the chip where the memory is to be read from
        :param int base_address:
            The address in SDRAM where the region of memory to be read starts
        :param int length: The length of the data to be read in bytes
        :param int cpu:
            the core ID used to read the memory of; should usually be 0 when
            reading from SDRAM, but may be other values when reading from DTCM.
        :return: A bytearray of data read
        :rtype: bytes
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If one of `x`, `y`, `cpu`, `base_address` or `length` is invalid
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def read_word(
            self, x: int, y: int, base_address: int, cpu: int = 0) -> int:
        """
        Read a word (usually of SDRAM) from the board.

        :param int x:
            The x-coordinate of the chip where the word is to be read from
        :param int y:
            The y-coordinate of the chip where the word is to be read from
        :param int base_address:
            The address (usually in SDRAM) where the word to be read starts
        :param int cpu:
            the core ID used to read the word; should usually be 0 when reading
            from SDRAM, but may be other values when reading from DTCM.
        :return: The unsigned integer value at ``base_address``
        :rtype: int
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If one of `x`, `y`, `cpu` or `base_address` is invalid
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def stop_application(self, app_id: int):
        """
        Sends a stop request for an app_id.

        :param int app_id: The ID of the application to send to
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If app_id is not a valid application ID
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def wait_for_cores_to_be_in_state(
            self, all_core_subsets: CoreSubsets, app_id: int,
            cpu_states: Union[CPUState, Iterable[CPUState]], *,
            timeout: Optional[float] = None,
            time_between_polls: float = 0.1,
            error_states: FrozenSet[CPUState] = frozenset((
                CPUState.RUN_TIME_EXCEPTION, CPUState.WATCHDOG)),
            counts_between_full_check: int = 100,
            progress_bar: Optional[ProgressBar] = None):
        """
        Waits for the specified cores running the given application to be
        in some target state or states. Handles failures.

        :param ~spinn_machine.CoreSubsets all_core_subsets:
            the cores to check are in a given sync state
        :param int app_id: the application ID that being used by the simulation
        :param cpu_states:
            The expected states once the applications are ready; success is
            when each application is in one of these states
        :type cpu_states: CPUState or iterable(CPUState)
        :param float timeout:
            The amount of time to wait in seconds for the cores to reach one
            of the states
        :param float time_between_polls: Time between checking the state
        :param set(CPUState) error_states:
            Set of states that the application can be in that indicate an
            error, and so should raise an exception
        :param int counts_between_full_check:
            The number of times to use the count signal before instead using
            the full CPU state check
        :param progress_bar: Possible progress bar to update.
        :type progress_bar: ~spinn_utilities.progress_bar.ProgressBar or None
        :raise SpinnmanTimeoutException:
            If a timeout is specified and exceeded.
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def send_signal(self, app_id: int, signal: Signal):
        """
        Send a signal to an application.

        :param int app_id: The ID of the application to send to
        :param ~spinnman.messages.scp.enums.Signal signal: The signal to send
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If signal is not a valid signal
            * If app_id is not a valid application ID
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        # Known usages are to send Signal START, SYNC0 and SYNC1
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def set_ip_tag(self, ip_tag: IPTag, use_sender: bool = False):
        """
        Set up an IP tag.

        :param ~spinn_machine.tags.IPTag ip_tag:
            The tag to set up.

            .. note::
                `board_address` can be `None`, in which case, the tag will be
                assigned to all boards.
        :param bool use_sender:
            Optionally use the sender host and port instead of
            the given host and port in the tag
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If the IP tag fields are incorrect
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def set_reverse_ip_tag(self, reverse_ip_tag: ReverseIPTag):
        """
        Set up a reverse IP tag.

        :param ~spinn_machine.tags.ReverseIPTag reverse_ip_tag:
            The reverse tag to set up.

            .. note::
                The `board_address` field can be `None`, in which case, the tag
                will be assigned to all boards.
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If the reverse IP tag fields are incorrect
            * If a packet is received that has invalid parameters
            * If the UDP port is one that is already used by SpiNNaker for
                system functions
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def clear_ip_tag(self, tag: int, board_address: Optional[str] = None):
        """
        Clear the setting of an IP tag.

        :param int tag: The tag ID
        :param str board_address:
            Board address where the tag should be cleared.
            If not specified, all AbstractSCPConnection connections will send
            the message to clear the tag
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If the tag is not a valid tag
            * If the connection cannot send SDP messages
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def get_tags(self, connection: Optional[SCAMPConnection] = None
                 ) -> Iterable[AbstractTag]:
        """
        Get the current set of tags that have been set on the board.

        :param AbstractSCPConnection connection:
            Connection from which the tags should be received.
            If not specified, all AbstractSCPConnection connections will be
            queried and the response will be combined.
        :return: An iterable of tags
        :rtype: iterable(~spinn_machine.tags.AbstractTag)
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If the connection cannot send SDP messages
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def malloc_sdram(
            self, x: int, y: int, size: int, app_id: int, tag: int = 0) -> int:
        """
        Allocates a chunk of SDRAM on a chip on the machine.

        :param int x: The x-coordinate of the chip onto which to ask for memory
        :param int y: The y-coordinate of the chip onto which to ask for memory
        :param int size: the amount of memory to allocate in bytes
        :param int app_id: The ID of the application with which to associate
            the routes.  If not specified, defaults to 0.
        :param int tag: the tag for the SDRAM, a 8-bit (chip-wide) tag that can
            be looked up by a SpiNNaker application to discover the address of
            the allocated block. If `0` then no tag is applied.
        :return: the base address of the allocated memory
        :rtype: int
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def load_multicast_routes(
            self, x: int, y: int, routes: Collection[MulticastRoutingEntry],
            app_id: int):
        """
        Load a set of multicast routes on to a chip.

        :param int x:
            The x-coordinate of the chip onto which to load the routes
        :param int y:
            The y-coordinate of the chip onto which to load the routes
        :param iterable(~spinn_machine.MulticastRoutingEntry) routes:
            An iterable of multicast routes to load
        :param int app_id: The ID of the application with which to associate
            the routes.  If not specified, defaults to 0.
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If any of the routes are invalid
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def load_fixed_route(
            self, x: int, y: int, fixed_route: RoutingEntry, app_id: int):
        """
        Loads a fixed route routing table entry onto a chip's router.

        :param int x:
            The x-coordinate of the chip onto which to load the routes
        :param int y:
            The y-coordinate of the chip onto which to load the routes
        :param ~spinn_machine.RoutingEntry fixed_route:
            the route for the fixed route entry on this chip
        :param int app_id: The ID of the application with which to associate
            the routes.  If not specified, defaults to 0.
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If any of the routes are invalid
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def read_fixed_route(self, x: int, y: int, app_id: int) -> RoutingEntry:
        """
        Reads a fixed route routing table entry from a chip's router.

        :param int x:
            The x-coordinate of the chip onto which to load the routes
        :param int y:
            The y-coordinate of the chip onto which to load the routes
        :param int app_id:
            The ID of the application with which to associate the
            routes.  If not specified, defaults to 0.
        :return: the route as a fixed route entry
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def get_multicast_routes(
            self, x: int, y: int,
            app_id: Optional[int] = None) -> List[MulticastRoutingEntry]:
        """
        Get the current multicast routes set up on a chip.

        :param int x:
            The x-coordinate of the chip from which to get the routes
        :param int y:
            The y-coordinate of the chip from which to get the routes
        :param int app_id:
            The ID of the application to filter the routes for. If
            not specified, will return all routes
        :return: An iterable of multicast routes
        :rtype: list(~spinn_machine.MulticastRoutingEntry)
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def clear_multicast_routes(self, x: int, y: int):
        """
        Remove all the multicast routes on a chip.

        :param int x: The x-coordinate of the chip on which to clear the routes
        :param int y: The y-coordinate of the chip on which to clear the routes
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def get_router_diagnostics(self, x: int, y: int) -> RouterDiagnostics:
        """
        Get router diagnostic information from a chip.

        :param int x:
            The x-coordinate of the chip from which to get the information
        :param int y:
            The y-coordinate of the chip from which to get the information
        :return: The router diagnostic information
        :rtype: RouterDiagnostics
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def get_scamp_connection_selector(self) -> MostDirectConnectionSelector:
        """
        Returns the most direct scamp connections

        :rtype: MostDirectConnectionSelecto
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def set_router_diagnostic_filter(
            self, x: int, y: int, position: int,
            diagnostic_filter: DiagnosticFilter):
        """
        Sets a router diagnostic filter in a router.

        :param int x:
            The X address of the router in which this filter is being set.
        :param int y:
            The Y address of the router in which this filter is being set.
        :param int position:
            The position in the list of filters where this filter is to be
            added.
        :param ~spinnman.model.DiagnosticFilter diagnostic_filter:
            The diagnostic filter being set in the placed, between 0 and 15.

            .. note::
                Positions 0 to 11 are used by the default filters,
                and setting these positions will result in a warning.
        :raise SpinnmanIOException:
            * If there is an error communicating with the board
            * If there is an error reading the data
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If x, y does not lead to a valid chip
            * If position is less than 0 or more than 15
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def clear_router_diagnostic_counters(self, x: int, y: int):
        """
        Clear router diagnostic information on a chip.

        :param int x: The x-coordinate of the chip
        :param int y: The y-coordinate of the chip
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            If a packet is received that has invalid parameters or a counter
            ID is out of range
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def close(self) -> None:
        """
        Close the transceiver and any threads that are running.
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def control_sync(self, do_sync: bool):
        """
        Control the synchronisation of the chips.

        :param bool do_sync: Whether to synchronise or not
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def update_provenance_and_exit(self, x: int, y: int, p: int):
        """
        Sends a command to update provenance and exit

        :param int x:
            The x-coordinate of the core
        :param int y:
            The y-coordinate of the core
        :param int p:
            The processor on the core
        """
        raise NotImplementedError("abstractmethod")

    @abstractmethod
    def send_chip_update_provenance_and_exit(self, x: int, y: int, p: int):
        """
        Sends a signal to update the provenance and exit

        :param int x:
        :param int y:
        :param int p:
        """
        raise NotImplementedError("abstractmethod")



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from spinn_utilities.overrides import overrides
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, BMPRequest, BMPResponse)
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.scp import SCPRequestHeader

_ONE_WORD = struct.Struct("<I")


class ReadFPGARegister(BMPRequest['_SCPReadFPGARegisterResponse']):
    """
    Requests the data from a FPGA's register.
    """
    __slots__ = ()

    def __init__(self, fpga_num: int, register: int, board: int):
        """
        Sets up a read FPGA register request.

        :param int fpga_num: FPGA number (0, 1 or 2) to communicate with.
        :param int register:
            Register address to read to (will be rounded down to
            the nearest 32-bit word boundary).
        :param int board: which board to request the FPGA register from
        """
        # check to stop people asking for non-word aligned memory addresses
        # inverses all bits of a value, so is basically a inverse mask for the
        # value entered.
        arg1 = register & (~0x3)
        super().__init__(
            board,
            SCPRequestHeader(command=SCPCommand.CMD_LINK_READ),
            argument_1=arg1, argument_2=4, argument_3=fpga_num)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> '_SCPReadFPGARegisterResponse':
        return _SCPReadFPGARegisterResponse(
            "Read FPGA register", SCPCommand.CMD_LINK_READ)


class _SCPReadFPGARegisterResponse(BMPResponse[int]):
    """
    An SCP response to a request for the version of software running.
    """
    def _parse_payload(self, data: bytes, offset: int) -> int:
        return _ONE_WORD.unpack_from(data, offset)[0]

    @property
    def fpga_register(self) -> int:
        """
        The register information received.

        :rtype: int
        """
        return self._value



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Iterable, Optional
from spinn_utilities.overrides import overrides
from spinnman.exceptions import SpinnmanInvalidParameterException
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.sdp import SDPFlag, SDPHeader
from spinnman.messages.scp.impl.check_ok_response import CheckOKResponse


_WAIT_FLAG = 0x1 << 18


class AppCopyRun(AbstractSCPRequest[CheckOKResponse]):
    """
    An SCP request to copy an application and start it.
    """
    __slots__ = "__link",

    def __init__(self, x: int, y: int, link: int, size: int, app_id: int,
                 processors: Optional[Iterable[int]],
                 checksum: int, wait: bool = False):
        """
        :param int x:
            The x-coordinate of the chip to read from, between 0 and 255
        :param int y:
            The y-coordinate of the chip to read from, between 0 and 255
        :param int link: The ID of the link from which to copy
        :param int size: The number of bytes to read, must be divisible by 4
        :param int app_id: The app to associate the copied binary with
        :param list(int) processors: The processors to start on the chip
        :param int checksum: The checksum of the data to copy
        :param bool wait: Whether to start in wait mode or not
        """
        # pylint: disable=too-many-arguments
        if size % 4 != 0:
            raise SpinnmanInvalidParameterException(
                "size", size, "The size must be a multiple of 4")

        processor_mask = 0
        if processors is not None:
            for processor in processors:
                processor_mask |= (1 << processor)

        processor_mask |= (app_id << 24)
        if wait:
            processor_mask |= _WAIT_FLAG
        self.__link = link

        arg1 = ((checksum & 0x1FFFFFFF) << 3) | link

        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_APP_COPY_RUN),
            argument_1=arg1, argument_2=size, argument_3=processor_mask)

    def __repr__(self):
        return f"{super(AppCopyRun, self).__repr__()} (Link {self.__link})"

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> CheckOKResponse:
        return CheckOKResponse(
            f"Application Copy Run (Link {self.__link})",
            SCPCommand.CMD_APP_COPY_RUN)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import struct
from typing import Generic, Optional, TypeVar, TYPE_CHECKING
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from .scp_response import AbstractSCPResponse
if TYPE_CHECKING:
    from spinnman.messages.sdp import SDPHeader
    from spinnman.messages.scp.scp_request_header import SCPRequestHeader
#: :meta private:
R = TypeVar("R", bound=AbstractSCPResponse)

_THREE_WORDS = struct.Struct("<III")


class AbstractSCPRequest(Generic[R], metaclass=AbstractBase):
    """
    Represents an Abstract SCP Request.
    """
    __slots__ = (
        "_argument_1",
        "_argument_2",
        "_argument_3",
        "_data",
        "_scp_request_header",
        "_sdp_header")

    DEFAULT_DEST_X_COORD = 255
    DEFAULT_DEST_Y_COORD = 255

    def __init__(
            self, sdp_header: SDPHeader, scp_request_header: SCPRequestHeader,
            argument_1: Optional[int] = None,
            argument_2: Optional[int] = None,
            argument_3: Optional[int] = None,
            data: Optional[bytes] = None):
        """
        :param SDPHeader sdp_header: The SDP header of the request
        :param SCPRequestHeader scp_request_header:
            The SCP header of the request
        :param int argument_1:
            The first argument, or `None` if no first argument
        :param int argument_2:
            The second argument, or `None` if no second argument
        :param int argument_3:
            The third argument, or `None` if no third argument
        :param data: The optional data, or `None` if no data
        :type data: bytearray or bytes or None
        """
        # pylint: disable=too-many-arguments
        self._sdp_header = sdp_header
        self._scp_request_header = scp_request_header
        self._argument_1 = argument_1
        self._argument_2 = argument_2
        self._argument_3 = argument_3
        self._data = data

    @property
    def sdp_header(self) -> SDPHeader:
        """
        The SDP header of the message.

        :rtype: SDPHeader
        """
        return self._sdp_header

    @property
    def scp_request_header(self) -> SCPRequestHeader:
        """
        The SCP request header of the message.

        :rtype: SCPRequestHeader
        """
        return self._scp_request_header

    @property
    def argument_1(self) -> Optional[int]:
        """
        The first argument, or `None` if no first argument.

        :rtype: int
        """
        return self._argument_1

    @property
    def argument_2(self) -> Optional[int]:
        """
        The second argument, or `None` if no second argument.

        :rtype: int
        """
        return self._argument_2

    @property
    def argument_3(self) -> Optional[int]:
        """
        The third argument, or `None` if no third argument.

        :rtype: int
        """
        return self._argument_3

    @property
    def data(self) -> Optional[bytes]:
        """
        The data, or `None` if no data.

        :rtype: bytearray
        """
        return self._data

    @property
    def bytestring(self) -> bytes:
        """
        The request as a byte-string.

        :rtype: bytes
        """
        data = (self._sdp_header.bytestring +
                self._scp_request_header.bytestring)
        data += _THREE_WORDS.pack(
            0 if self._argument_1 is None else self._argument_1,
            0 if self._argument_2 is None else self._argument_2,
            0 if self._argument_3 is None else self._argument_3)
        if self._data is not None:
            data += bytes(self._data)
        return data

    def __repr__(self):
        # Default is to return just the command, but can be overridden
        return str(self._scp_request_header.command)

    def __str__(self):
        return self.__repr__()

    @abstractmethod
    def get_scp_response(self) -> R:
        """
        Get an SCP response message to be used to process any response
        received.

        :return: An SCP response, or `None` if no response is required
        :rtype: AbstractSCPResponse
        """
        raise NotImplementedError



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class P2PTableRoute(Enum):
    """
    P2P Routing table routes.
    """
    EAST = 0b000
    NORTH_EAST = 0b001
    NORTH = 0b010
    WEST = 0b011
    SOUTH_WEST = 0b100
    SOUTH = 0b101
    #: No route to this chip
    NONE = 0b110
    #: Route to the monitor on the current chip
    MONITOR = 0b111



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, BMPRequest, BMPResponse)
from spinnman.messages.scp.enums import BMPInfo, SCPCommand
from spinnman.model import ADCInfo


class ReadADC(BMPRequest['_SCPReadADCResponse']):
    """
    SCP Request for the data from the BMP including voltages and
    temperature.

    This class is currently deprecated and untested as there is no
    known use except for Transceiver.read_adc_data which is itself
    deprecated.

    .. note::
        The equivalent code in Java is *not* deprecated.
    """
    __slots__ = ()

    def __init__(self, board: int):
        """
        :param int board: which board to request the ADC data from
        """
        super().__init__(
            board,
            SCPRequestHeader(command=SCPCommand.CMD_BMP_INFO),
            argument_1=BMPInfo.ADC.value)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> '_SCPReadADCResponse':
        return _SCPReadADCResponse()


class _SCPReadADCResponse(BMPResponse[ADCInfo]):
    """
    An SCP response to a request for ADC information.
    """
    __slots__ = ("_adc_info", )

    def __init__(self) -> None:
        super().__init__("Read ADC", SCPCommand.CMD_BMP_INFO)
        self._adc_info = None

    def _parse_payload(self, data: bytes, offset: int) -> ADCInfo:
        return ADCInfo(data, offset)

    @property
    def adc_info(self):
        """
        The ADC information.
        """
        return self._value



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinnman.constants import EIEIO_COMMAND_IDS
from .eieio_command_message import EIEIOCommandMessage
from .eieio_command_header import EIEIOCommandHeader


class NotificationProtocolStartResume(EIEIOCommandMessage):
    """
    Packet which indicates that the toolchain has started or resumed.

    This message is not sent to SpiNNaker boards but rather to an auxiliary
    tool (e.g., data visualiser).
    """
    __slots__ = ()

    def __init__(self):
        super().__init__(EIEIOCommandHeader(
            EIEIO_COMMAND_IDS.START_RESUME_NOTIFICATION))

    @staticmethod
    def from_bytestring(command_header, data, offset):
        return NotificationProtocolStartResume()



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class AllocFree(Enum):
    """
    The SCP Allocation and Free codes.
    """
    #: Allocate SDRAM.
    ALLOC_SDRAM = 0
    #: Free SDRAM using a Pointer.
    FREE_SDRAM_BY_POINTER = 1
    #: Free SDRAM using am APP ID.
    FREE_SDRAM_BY_APP_ID = 2
    #: Allocate Routing Entries.
    ALLOC_ROUTING = 3
    #: Free Routing Entries by Pointer.
    FREE_ROUTING_BY_POINTER = 4
    #: Free Routing Entries by APP ID.
    FREE_ROUTING_BY_APP_ID = 5



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from spinnman.messages.scp.impl import GetVersion
from spinnman.constants import N_RETRIES
from spinnman.model import VersionInfo
from spinnman.messages.scp.impl.get_version_response import GetVersionResponse

from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)


class GetVersionProcess(AbstractMultiConnectionProcess[GetVersionResponse]):
    """
    A process for getting the version of the machine.
    """
    __slots__ = "_version_info",

    def __init__(self, connection_selector: ConnectionSelector,
                 n_retries: int = N_RETRIES):
        """
        :param ConnectionSelector connection_selector:
        """
        super().__init__(connection_selector, n_retries)
        self._version_info: Optional[VersionInfo] = None

    def _get_response(self, version_response: GetVersionResponse):
        """
        :param GetVersionResponse version_response:
        """
        self._version_info = version_response.version_info

    def get_version(self, x: int, y: int, p: int) -> VersionInfo:
        """
        :param int x:
        :param int y:
        :param int p:
        :rtype: VersionInfo
        """
        with self._collect_responses():
            self._send_request(GetVersion(x=x, y=y, p=p), self._get_response)
        assert self._version_info is not None
        return self._version_info



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array
import math
import os
import time
from typing import Any, Dict, Iterable, Optional, Tuple

from spinnman.data import SpiNNManDataView
from spinnman.exceptions import SpinnmanIOException

from .system_variable_boot_values import (
    SystemVariableBootValues, SystemVariableDefinition)
from .spinnaker_boot_message import SpinnakerBootMessage
from .spinnaker_boot_op_code import SpinnakerBootOpCode

_BOOT_MESSAGE_DATA_WORDS = 256
_BOOT_MESSAGE_DATA_BYTES = _BOOT_MESSAGE_DATA_WORDS * 4
_BOOT_IMAGE_MAX_BYTES = 32 * 1024
_BOOT_DATA_FILE_NAME = "scamp.boot"
_BOOT_STRUCT_REPLACE_OFFSET = 384 // 4
_BOOT_STRUCT_REPLACE_LENGTH = 128 // 4
_BOOT_DATA_OPERAND_1 = ((_BOOT_MESSAGE_DATA_BYTES // 4) - 1) << 8


class SpinnakerBootMessages(object):
    """
    A set of boot messages to be sent to boot the board.
    """
    __slots__ = (
        "_boot_data",
        "_n_bytes_to_read",
        "_no_data_packets")

    def __init__(self, extra_boot_values: Optional[Dict[
            SystemVariableDefinition, Any]] = None):
        """
        :param extra_boot_values:
            Any additional or overwrite values to set during boot.
            This should only be used for values which are not standard
            based on the board version.
            for example this may include an Led_0 value.
        :type extra_boot_values: dict(SystemVariableDefinition, object)
        :raise SpinnmanInvalidParameterException:
            If the board version is not supported
        :raise SpinnmanIOException:
            If there is an error assembling the packets
        """
        version = SpiNNManDataView.get_machine_version()

        # Get the boot packet values
        spinnaker_boot_value = SystemVariableBootValues()

        spinnaker_boot_value.set_value(
            SystemVariableDefinition.hardware_version, version.number)

        current_time = int(time.time())
        spinnaker_boot_value.set_value(
            SystemVariableDefinition.unix_timestamp, current_time)
        spinnaker_boot_value.set_value(
            SystemVariableDefinition.boot_signature, current_time)
        spinnaker_boot_value.set_value(
            SystemVariableDefinition.is_root_chip, 1)

        # Set any additional values
        if extra_boot_values is not None:
            for variable, value in extra_boot_values.items():
                spinnaker_boot_value.set_value(variable, value)

        # Get the data as an array, to be used later
        spinnaker_boot_data = array.array("I", spinnaker_boot_value.bytestring)

        # Find the data file and size
        boot_data_file, boot_data_size = self._get_boot_image_file()

        # Compute how many packets to send
        n_words_to_read = boot_data_size // 4
        self._no_data_packets = int(math.ceil(
            float(boot_data_size) / _BOOT_MESSAGE_DATA_BYTES))

        # Read the data
        boot_data = array.array("I")
        with open(boot_data_file, "rb") as f:
            boot_data.fromfile(f, n_words_to_read)

        # Replace the appropriate part with the custom boot options
        offset = _BOOT_STRUCT_REPLACE_OFFSET
        length = _BOOT_STRUCT_REPLACE_LENGTH
        boot_data[offset:offset + length] = spinnaker_boot_data[0:length]

        # Byte swap and store the data for later use
        boot_data.byteswap()
        self._boot_data = boot_data.tobytes()
        self._n_bytes_to_read = n_words_to_read * 4

    @staticmethod
    def _get_boot_image_file() -> Tuple[str, int]:
        """
        :rtype: tuple(str,int)
        """
        this_dir, _ = os.path.split(__file__)
        file_name = os.path.join(this_dir, "boot_data", _BOOT_DATA_FILE_NAME)
        file_size = os.stat(file_name).st_size
        if file_size > _BOOT_IMAGE_MAX_BYTES:
            raise SpinnmanIOException(
                f"The boot file is too big at {file_size} bytes "
                "(only files up to 32KB are acceptable")
        elif file_size % 4 != 0:
            raise SpinnmanIOException(
                f"The boot file size of {file_size} bytes "
                "must be divisible by 4")
        return file_name, file_size

    def _get_packet_data(self, block_id: int) -> bytes:
        """
        Read a packet of data.

        :param int block_id:
        :rtype: bytes
        """
        offset = block_id * _BOOT_MESSAGE_DATA_BYTES
        n_bytes = min(self._n_bytes_to_read - offset, _BOOT_MESSAGE_DATA_BYTES)
        return self._boot_data[offset:offset + n_bytes]

    @property
    def messages(self) -> Iterable[SpinnakerBootMessage]:
        """
        An iterable of message to be sent.

        :rtype: iterable(SpinnakerBootMessage)
        """
        # Construct and yield the start packet
        yield SpinnakerBootMessage(
            opcode=SpinnakerBootOpCode.FLOOD_FILL_START,
            operand_1=0, operand_2=0, operand_3=self._no_data_packets - 1)

        # Construct and yield the data packets
        for block_id in range(0, self._no_data_packets):
            operand_1 = _BOOT_DATA_OPERAND_1 | (block_id & 0xFF)
            yield SpinnakerBootMessage(
                opcode=SpinnakerBootOpCode.FLOOD_FILL_BLOCK,
                operand_1=operand_1, operand_2=0, operand_3=0,
                data=self._get_packet_data(block_id))

        # Construct and yield the end packet
        yield SpinnakerBootMessage(
            opcode=SpinnakerBootOpCode.FLOOD_FILL_CONTROL,
            operand_1=1, operand_2=0, operand_3=0)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
import struct
from typing import Union
from typing_extensions import TypeAlias
from spinnman.messages.scp.enums import SCPCommand

_TWO_SHORTS = struct.Struct("<2H")
_Command: TypeAlias = Union[SCPCommand, Enum]


class SCPRequestHeader(object):
    """
    Represents the header of an SCP Request.
    """
    __slots__ = (
        "_command",
        "_sequence")

    def __init__(self, command: _Command, sequence: int = 0):
        """
        :param SCPCommand command: The SCP command
        :param int sequence:
            The number of the SCP packet in order of all packets
            sent or received, between 0 and 65535
        """
        self._command = command
        self._sequence = sequence

    @property
    def command(self) -> _Command:
        """
        The command of the SCP packet.

        :rtype: SCPCommand
        """
        return self._command

    @property
    def sequence(self) -> int:
        """
        The sequence number of the SCP packet, between 0 and 65535.

        :rtype: int
        """
        return self._sequence

    @sequence.setter
    def sequence(self, sequence: int):
        """
        Set the sequence number of the SCP packet.

        :param int sequence: The sequence number to set, between 0 and 65535
        """
        self._sequence = sequence

    @property
    def bytestring(self) -> bytes:
        """
        The header as a byte-string.

        :rtype: bytes
        """
        return _TWO_SHORTS.pack(self._command.value, self._sequence)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
from spinnman.model import BMPConnectionData
from spinnman.messages.scp.impl import IPTagSet
from spinnman.messages.sdp import SDPMessage, SDPHeader, SDPFlag
from spinnman.constants import SCP_SCAMP_PORT, CPU_INFO_BYTES, CPU_INFO_OFFSET
from spinnman.connections.udp_packet_connections import (
    SCAMPConnection, UDPConnection)
from spinnman.exceptions import SpinnmanTimeoutException


def work_out_bmp_from_machine_details(
        hostname: str) -> BMPConnectionData:
    """
    Work out the BMP connection IP address given the machine details.
    This is assumed to be the IP address of the machine, with 1 subtracted
    from the final part e.g. if the machine IP address is 192.168.0.5, the
    BMP IP address is assumed to be 192.168.0.4

    :param str hostname: the SpiNNaker machine main hostname or IP address
    :return: The BMP connection data
    :rtype: BMPConnectionData
    """
    # take the IP address, split by dots, and subtract 1 off last bit
    ip_bits = socket.gethostbyname(hostname).split(".")
    ip_bits[-1] = str(int(ip_bits[-1]) - 1)
    bmp_ip_address = ".".join(ip_bits)

    # add board scope for each split
    # if None, the end user didn't enter anything, so assume one board
    # starting at position 0
    board_range = [0]

    # Assume a single board with no cabinet or frame specified
    return BMPConnectionData(ip_address=bmp_ip_address,
                             boards=board_range, port_num=SCP_SCAMP_PORT)


def get_vcpu_address(p: int) -> int:
    """
    Get the address of the vcpu_t structure for the given core.

    :param int p: The core
    :rtype: int
    """
    return CPU_INFO_OFFSET + (CPU_INFO_BYTES * p)


def send_port_trigger_message(
        connection: UDPConnection, board_address: str):
    """
    Sends a port trigger message using a connection to (hopefully) open a
    port in a NAT and/or firewall to allow incoming packets to be received.

    :param UDPConnection connection:
        The UDP connection down which the trigger message should be sent
    :param str board_address:
        The IP address of the SpiNNaker board to which the message should be
        sent
    """

    # Set up the message so that no reply is expected and it is sent to an
    # invalid port for SCAMP.  The current version of SCAMP will reject
    # this message, but then fail to send a response since the
    # REPLY_NOT_EXPECTED flag is set (see scamp-3.c line 728 and 625-644)
    trigger_message = SDPMessage(SDPHeader(
        flags=SDPFlag.REPLY_NOT_EXPECTED, tag=0, destination_port=3,
        destination_cpu=0, destination_chip_x=0, destination_chip_y=0))
    trigger_message.sdp_header.update_for_send(0, 0)
    connection.send_to(
        trigger_message.bytestring, (board_address, SCP_SCAMP_PORT))


def reprogram_tag(connection: SCAMPConnection, tag: int, strip: bool = True):
    """
    Reprogram an IP Tag to send responses to a given SCAMPConnection.

    :param SCAMPConnection connection: The connection to target the tag at
    :param int tag: The id of the tag to set
    :param bool strip:
        True if the tag should strip SDP headers from outgoing messages
    :raises SpinnmanTimeoutException:
        If things time out several times
    """
    request = IPTagSet(
        connection.chip_x, connection.chip_y, [0, 0, 0, 0], 0, tag,
        strip=strip, use_sender=True)
    data = connection.get_scp_data(request)
    exception = None
    for _ in range(3):
        try:
            connection.send(data)
            _, _, response, offset = connection.receive_scp_response()
            request.get_scp_response().read_bytestring(response, offset)
            return
        except SpinnmanTimeoutException as e:
            exception = e
    # Should be impossible to get here with Exception is None
    raise exception or Exception


def reprogram_tag_to_listener(
        connection: UDPConnection, x: int, y: int, ip_address: str, tag: int,
        strip: bool = True, read_response: bool = True):
    """
    Reprogram an IP Tag to send responses to a given connection that is
    not connected to a specific board. Such connections are normally
    receive-only connections.

    :param UDPConnection connection: The connection to target the tag at
    :param int x:
        The X coordinate of the Ethernet-enabled chip that should send to the
        connection
    :param int y:
        The Y coordinate of the Ethernet-enabled chip that should send to the
        connection
    :param str ip_address:
        The IP address of the Ethernet-enabled chip that should be given the
        message
    :param int tag: The id of the tag to set
    :param bool strip:
        True if the tag should strip SDP headers from outgoing messages
    :param bool read_response:
        True if the response to the reprogramming should be read
    :raises SpinnmanTimeoutException:
        If things time out several times
    """
    request = IPTagSet(
        x, y, [0, 0, 0, 0], 0, tag,
        strip=strip, use_sender=True)
    request.sdp_header.update_for_send(x, y)
    send_data = b'\0\0' + request.bytestring
    exception = None
    for _ in range(3):
        try:
            connection.send_to(send_data, (ip_address, SCP_SCAMP_PORT))
            if read_response:
                request.get_scp_response().read_bytestring(
                    connection.receive(), 2)
            return
        except SpinnmanTimeoutException as e:
            exception = e
    # Should be impossible to get here with exception is None
    raise exception or Exception



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinnman.messages.eieio.eieio_type import EIEIOType


class AbstractDataElement(object, metaclass=AbstractBase):
    """
    A marker interface for possible data elements in the EIEIO data packet.
    """

    __slots__ = ()

    @abstractmethod
    def get_bytestring(self, eieio_type: EIEIOType) -> bytes:
        """
        Get a byte-string for the given type

        :param EIEIOType eieio_type: The type of the message being written
        :return: A byte-string for the element
        :rtype: bytes
        :raise SpinnmanInvalidParameterException:
            If the type is incompatible with the element
        """
        raise NotImplementedError



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import IntEnum


class ProxyProtocol(IntEnum):
    """
    Websocket binary frame type identifiers in the Spalloc Proxy protocol.
    """
    #: Message relating to opening a channel
    OPEN = 0
    #: Message relating to closing a channel
    CLOSE = 1
    #: Message sent on a channel
    MSG = 2
    #: Message relating to opening an unbound listen-only channel
    OPEN_UNBOUND = 3
    #: Message sent on an unbound channel to a given board
    MSG_TO = 4
    #: Message relating to an error when opening or closing a channel
    ERROR = 5



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class LEDAction(Enum):
    """
    The SCP LED actions.
    """
    #: Toggle the LED status.
    TOGGLE = 1
    #: Turn the LED off.
    OFF = 2
    #: Turn the LED on.
    ON = 3



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Generic, Iterable, Optional, TypeVar, Union
from typing_extensions import TypeAlias

from spinnman.messages.sdp import SDPFlag, SDPHeader
from spinnman.messages.scp import SCPRequestHeader

from .bmp_response import BMPResponse
from .scp_request import AbstractSCPRequest

#: The type of boards parameters.
Boards: TypeAlias = Union[int, Iterable[int]]
R = TypeVar("R", bound=BMPResponse)


class BMPRequest(  # pylint: disable=abstract-method
        AbstractSCPRequest[R], Generic[R]):
    """
    An SCP request intended to be sent to a BMP.
    """
    __slots__ = ()

    def __init__(self, boards: Boards,
                 scp_request_header: SCPRequestHeader,
                 argument_1: Optional[int] = None,
                 argument_2: Optional[int] = None,
                 argument_3: Optional[int] = None,
                 data: Optional[bytes] = None):
        """
        :param boards: The board or boards to be addressed by this request
        :type boards: int or list(int) or tuple(int)
        :param SCPRequestHeader scp_request_header: The SCP request header
        :param int argument_1: The optional first argument
        :param int argument_2: The optional second argument
        :param int argument_3: The optional third argument
        :param bytes data: The optional data to be sent
        """
        # pylint: disable=too-many-arguments
        sdp_header = SDPHeader(
            flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
            destination_cpu=BMPRequest.get_first_board(boards),
            destination_chip_x=0, destination_chip_y=0)
        super().__init__(
            sdp_header, scp_request_header,
            argument_1, argument_2, argument_3, data)

    @staticmethod
    def get_first_board(boards: Boards) -> int:
        """
        Get the first board ID given a board ID or collection of board IDs.
        """
        if isinstance(boards, int):
            return boards
        return min(boards)

    @staticmethod
    def get_board_mask(boards: Boards) -> int:
        """
        Get the board mask given a board ID or collection of board IDs.
        """
        if isinstance(boards, int):
            return 1 << boards
        return sum(1 << board for board in boards)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class MailboxCommand(Enum):
    """
    Commands sent between an application and the monitor processor.
    """
    #: The mailbox is idle
    SHM_IDLE = 0
    #: The mailbox contains an SDP message
    SHM_MSG = 1
    #: The mailbox contains a no-operation (used for watchdog monitoring)
    SHM_NOP = 2
    #: The mailbox contains a signal
    SHM_SIGNAL = 3
    #: The mailbox contains a command
    SHM_CMD = 4



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, Generic, TypeVar
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
#: :meta private:
T = TypeVar("T")


# Should inherit from Connection, but doesn't for Multiple Inheritance reasons
class Listenable(Generic[T], metaclass=AbstractBase):
    """
    An interface for connections that can listen for incoming messages.

    Implementing this interface means that the connection can be used with
    :py:class:`ConnectionListener`.
    """

    __slots__ = ()

    @abstractmethod
    def get_receive_method(self) -> Callable[[], T]:
        """
        Get the method that receives for this connection.
        """
        raise NotImplementedError

    @abstractmethod
    def is_ready_to_receive(self, timeout: float = 0) -> bool:
        """
        Determines if there is an SCP packet to be read without blocking.

        :param int timeout:
            The time to wait before returning if the connection is not ready
        :return: True if there is an SCP packet to be read
        :rtype: bool
        """
        raise NotImplementedError



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .app_stop import AppStop
from .app_copy_run import AppCopyRun
from .application_run import ApplicationRun
from .bmp_get_version import BMPGetVersion
from .check_ok_response import CheckOKResponse
from .count_state import CountState
from .do_sync import DoSync
from .fill_request import FillRequest
from .flood_fill_data import FloodFillData
from .flood_fill_end import FloodFillEnd
from .flood_fill_start import FloodFillStart
from .get_chip_info import GetChipInfo
from .get_version import GetVersion
from .iptag_clear import IPTagClear
from .iptag_get import IPTagGet
from .iptag_get_info import IPTagGetInfo
from .iptag_set import IPTagSet
from .iptag_set_tto import IPTagSetTTO
from .read_fpga_register import ReadFPGARegister
from .read_link import ReadLink
from .read_memory import ReadMemory
from .reverse_iptag_set import ReverseIPTagSet
from .router_alloc import RouterAlloc
from .router_clear import RouterClear
from .router_init import RouterInit
from .sdram_alloc import SDRAMAlloc
from .sdram_de_alloc import SDRAMDeAlloc
from .send_signal import SendSignal
from .set_power import SetPower
from .write_fpga_register import WriteFPGARegister
from .write_link import WriteLink
from .write_memory import WriteMemory
from .fixed_route_init import FixedRouteInit
from .fixed_route_read import FixedRouteRead

__all__ = ["AppStop", "ApplicationRun", "AppCopyRun",
           "BMPGetVersion", "CheckOKResponse", "GetChipInfo", "CountState",
           "DoSync",
           "FloodFillData", "FillRequest",
           "FloodFillEnd", "FloodFillStart",
           "IPTagClear", "IPTagGet", "IPTagGetInfo",
           "IPTagSet", "IPTagSetTTO",
           "SetPower", "ReadFPGARegister", "ReadLink",
           "ReadMemory", "ReverseIPTagSet",
           "RouterAlloc", "RouterClear",
           "RouterInit", "SDRAMAlloc",
           "SDRAMDeAlloc", "SendSignal",
           "GetVersion", "WriteFPGARegister", "FixedRouteRead",
           "WriteLink", "WriteMemory", "FixedRouteInit"]



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Optional
from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, AbstractSCPResponse)
from spinnman.messages.scp.enums import (
    AllocFree, SCPCommand, SCPResult)
from spinnman.messages.sdp import SDPFlag, SDPHeader
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException

_ONE_WORD = struct.Struct("<I")


class RouterAllocResponse(AbstractSCPResponse):
    """
    An SCP response to a request to allocate router entries.
    """
    __slots__ = "_base_address",

    def __init__(self) -> None:
        super().__init__()
        self._base_address: Optional[int] = None

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data: bytes, offset: int):
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                "Router Allocation", "CMD_ALLOC", result.name)
        self._base_address = _ONE_WORD.unpack_from(data, offset)[0]

    @property
    def base_address(self) -> int:
        """
        The base address allocated, or 0 if none.

        :rtype: int
        """
        return self._base_address or 0


class RouterAlloc(AbstractSCPRequest[RouterAllocResponse]):
    """
    An SCP Request to allocate space for routing entries.
    """
    __slots__ = ()

    def __init__(self, x: int, y: int, app_id: int, n_entries: int):
        """
        :param int x:
            The x-coordinate of the chip to allocate on, between 0 and 255
        :param int y:
            The y-coordinate of the chip to allocate on, between 0 and 255
        :param int app_id: The ID of the application, between 0 and 255
        :param int n_entries: The number of entries to allocate
        """
        # pylint: disable=unsupported-binary-operation
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_ALLOC),
            argument_1=(
                (app_id << 8) |
                AllocFree.ALLOC_ROUTING.value),
            argument_2=n_entries)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> RouterAllocResponse:
        return RouterAllocResponse()



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import logging
from typing import BinaryIO, Optional, Tuple, Union, TYPE_CHECKING
from spinn_utilities.log import FormatAdapter
from spinn_machine.data import MachineDataView
from spinnman.utilities.appid_tracker import AppIdTracker
if TYPE_CHECKING:
    from spinnman.processes import MostDirectConnectionSelector
    from spinnman.transceiver import Transceiver

logger = FormatAdapter(logging.getLogger(__name__))
# pylint: disable=protected-access


class _SpiNNManDataModel(object):
    """
    Singleton data model.

    This class should not be accessed directly please use the DataView and
    DataWriter classes.
    Accessing or editing the data held here directly is *not supported!*

    There may be other DataModel classes which sit next to this one and hold
    additional data. The DataView and DataWriter classes will combine these
    as needed.

    What data is held where and how can change without notice.
    """

    __singleton = None

    __slots__ = [
        # Data values cached
        "_app_id",
        "_app_id_tracker",
        "_scamp_connection_selector",
        "_transceiver",
    ]

    def __new__(cls) -> _SpiNNManDataModel:
        if cls.__singleton:
            return cls.__singleton
        obj = object.__new__(cls)
        cls.__singleton = obj
        obj._transceiver = None
        obj._clear()
        return obj

    def _clear(self) -> None:
        """
        Clears out all data.
        """
        self._hard_reset()

    def _hard_reset(self) -> None:
        """
        Clears out all data that should change after a reset and graph change.
        """
        self._app_id: Optional[int] = None
        self._app_id_tracker: Optional[AppIdTracker] = None
        self._soft_reset()
        self._scamp_connection_selector: Optional[
            MostDirectConnectionSelector] = None
        if self._transceiver:
            try:
                self._transceiver.close()
            except Exception as ex:  # pylint: disable=broad-except
                logger.exception(
                    f"Error {ex} when closing the transceiver ignored")
        self._transceiver: Optional[Transceiver] = None

    def _soft_reset(self) -> None:
        """
        Clears timing and other data that should changed every reset.
        """
        # Holder for any later additions


class SpiNNManDataView(MachineDataView):
    """
    Adds the extra Methods to the View for SpiNNMan level.

    See :py:class:`~spinn_utilities.data.UtilsDataView` for a more detailed
    description.

    This class is designed to only be used directly within the SpiNNMan
    repository as all methods are available to subclasses
    """

    __data = _SpiNNManDataModel()
    __slots__ = ()

    # transceiver methods

    @classmethod
    def has_transceiver(cls) -> bool:
        """
        Reports if a transceiver is currently set.

        :rtype: bool
        """
        return cls.__data._transceiver is not None

    @classmethod
    def get_transceiver(cls) -> Transceiver:
        """
        The transceiver description.

        :rtype: ~spinnman.transceiver.Transceiver
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the transceiver is currently unavailable
        """
        if cls.__data._transceiver is None:
            raise cls._exception("transceiver")
        return cls.__data._transceiver

    @classmethod
    def read_memory(
            cls, x: int, y: int, base_address: int, length: int, *,
            cpu: int = 0) -> bytes:
        """
        Read some areas of memory (usually SDRAM) from the board.

        Syntactic sugar for `get_transceiver().read_memory()`.

        :param int x:
            The x-coordinate of the chip where the memory is to be read from
        :param int y:
            The y-coordinate of the chip where the memory is to be read from
        :param int base_address:
            The address in SDRAM where the region of memory to be read starts
        :param int length: The length of the data to be read in bytes
        :param int cpu:
            the core ID used to read the memory of; should usually be 0 when
            reading from SDRAM, but may be other values when reading from DTCM.
        :return: A bytearray of data read
        :rtype: bytes
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the transceiver is currently unavailable
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If one of `x`, `y`, `cpu`, `base_address` or `length` is invalid
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        try:
            return cls.get_transceiver().read_memory(
                x, y, base_address, length, cpu=cpu)
        except AttributeError as ex:
            raise cls._exception("transceiver") from ex

    @classmethod
    def write_memory(
            cls, x: int, y: int, base_address: int, data: Union[
                BinaryIO, bytes, bytearray, int, str], *,
            n_bytes: Optional[int] = None, offset: int = 0,
            cpu: int = 0) -> Tuple[int, int]:
        """
        Write to the SDRAM on the board.

        Syntactic sugar for `get_transceiver().write_memory()`.

        :param int x:
            The x-coordinate of the chip where the memory is to be written to
        :param int y:
            The y-coordinate of the chip where the memory is to be written to
        :param int base_address:
            The address in SDRAM where the region of memory is to be written
        :param data: The data to write.  Should be one of the following:

            * An instance of RawIOBase
            * A bytearray/bytes
            * A single integer - will be written in little-endian byte order
            * A filename of a data file
        :type data:
            ~io.RawIOBase or bytes or bytearray or int or str
        :param int n_bytes:
            The amount of data to be written in bytes.  If not specified:

            * If `data` is an RawIOBase, an error is raised
            * If `data` is a byte string (bytearray or bytes), the length of\
              the byte string will be used
            * If `data` is an int, 4 will be used
            * If `data` is a str, the length of the file will be used
        :param int offset: The offset from which the valid data begins
        :param int cpu: The optional CPU to write to
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the transceiver is currently unavailable
        :raise SpinnmanIOException:
            * If there is an error communicating with the board
            * If there is an error reading the data
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If `x, y` does not lead to a valid chip
            * If a packet is received that has invalid parameters
            * If `base_address` is not a positive integer
            * If `data` is an RawIOBase but `n_bytes` is not specified
            * If `data` is an int and `n_bytes` is more than 4
            * If `n_bytes` is less than 0
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        return cls.get_transceiver().write_memory(
            x, y, base_address, data,
            n_bytes=n_bytes, offset=offset, cpu=cpu)

    # app_id methods

    @classmethod
    def get_app_id(cls) -> int:
        """
        Gets the main app_id used by the transceiver.

        This method will create a new app_id if one has not yet been created.

        :rtype: int
        """
        if cls.__data._app_id is None:
            cls.__data._app_id = cls.get_new_id()
        return cls.__data._app_id

    @classmethod
    def get_new_id(cls) -> int:
        """
        Gets a new id from the current `app_id_tracker`

        previously `get_transceiver().app_id_tracker().get_new_id()`

        :rtype: AppIdTracker
        """
        if cls.__data._app_id_tracker is None:
            cls.__data._app_id_tracker = AppIdTracker()
        return cls.__data._app_id_tracker.get_new_id()

    @classmethod
    def free_id(cls, app_id: int):
        """
        Frees up an app_id.

        previously `get_transceiver().app_id_tracker().free_id(app_id)`

        :param int app_id:
        """
        if cls.__data._app_id_tracker:
            cls.__data._app_id_tracker.free_id(app_id)

    @classmethod
    def get_scamp_connection_selector(cls) -> MostDirectConnectionSelector:
        """
        Gets the SCAMP connection selector from the transceiver.

        Syntactic sugar for `get_transceiver().get_scamp_connection_selector()`

        :rtype: MostDirectConnectionSelector
        :raises ~spinn_utilities.exceptions.SpiNNUtilsException:
            If the transceiver is currently unavailable
        """
        if not cls.__data._scamp_connection_selector:
            cls.__data._scamp_connection_selector =\
                cls. get_transceiver().get_scamp_connection_selector()
        return cls.__data._scamp_connection_selector



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Optional

from spinnman.messages.eieio import EIEIOType, EIEIOPrefix
from spinnman.exceptions import SpinnmanInvalidPacketException

_PATTERN_BBHH = struct.Struct("<BBHH")
_PATTERN_BBH = struct.Struct("<BBH")
_PATTERN_BBHI = struct.Struct("<BBHI")
_PATTERN_BBI = struct.Struct("<BBI")
_PATTERN_BB = struct.Struct("<BB")
_PATTERN_HH = struct.Struct("<HH")
_PATTERN_H = struct.Struct("<H")
_PATTERN_HI = struct.Struct("<HI")
_PATTERN_I = struct.Struct("<I")


class EIEIODataHeader(object):
    """
    The header part of EIEIO data.
    """
    __slots__ = (
        "_count",
        "_eieio_type",
        "_is_time",
        "_payload_base",
        "_prefix",
        "_prefix_type",
        "_tag")

    def __init__(self, eieio_type: EIEIOType, tag: int = 0,
                 prefix: Optional[int] = None,
                 prefix_type: EIEIOPrefix = EIEIOPrefix.LOWER_HALF_WORD,
                 payload_base: Optional[int] = None, is_time: bool = False,
                 count: int = 0):
        """
        EIEIO header for data packets.

        :param EIEIOType eieio_type: the type of message
        :param int tag: the tag of the message (0 by default)
        :param prefix: the key prefix of the message or `None` if not prefixed
        :type prefix: int or None
        :param EIEIOPrefix prefix_type:
            the position of the prefix (upper or lower)
        :param payload_base:
            The base payload to be applied, or `None` if no base payload
        :type payload_base: int or None
        :param bool is_time:
            True if the payloads should be taken to be timestamps, or False
            otherwise
        :param int count: Count of the number of items in the packet
        """
        # pylint: disable=too-many-arguments
        self._eieio_type = eieio_type
        self._tag = tag
        self._prefix = prefix
        self._prefix_type = prefix_type
        self._payload_base = payload_base
        self._is_time = is_time
        self._count = count

    @property
    def eieio_type(self) -> EIEIOType:
        """
        Gets the eieio_type passed into the init.

        :rtype: EIEIOType
        """
        return self._eieio_type

    @property
    def tag(self) -> int:
        """
        Gets the tag value passed into the init.

        :rtype: int
        """
        return self._tag

    @property
    def prefix(self) -> Optional[int]:
        """
        Gets prefix passed into the init (if applicable).

        :rtype: int or None
        """
        return self._prefix

    @property
    def prefix_type(self) -> EIEIOPrefix:
        """
        Gets the prefix_type passed into the init.

        :rtype: EIEIOPrefix
        """
        return self._prefix_type

    @property
    def payload_base(self) -> Optional[int]:
        """
        Gets the payload_base value passed into the init (if applicable).

        :rtype: int or None
        """
        return self._payload_base

    @property
    def is_time(self) -> bool:
        """
        Gets the is_time value passed into the init.

        :rtype: bool
        """
        return self._is_time

    @property
    def count(self) -> int:
        """
        Count of the number of items in the packet

        :rtype: int
        """
        return self._count

    @count.setter
    def count(self, count: int):
        """
        Sets the Count of the number of items in the packet

        :param int count: the new value
        """
        self._count = count

    @property
    def size(self) -> int:
        """
        Get the size of a header with the given parameters.

        :rtype: int
        """
        return EIEIODataHeader.get_header_size(
            self._eieio_type, self._prefix is not None,
            self._payload_base is not None)

    @staticmethod
    def get_header_size(eieio_type, is_prefix=False, is_payload_base=False):
        """
        Get the size of a header with the given parameters.

        :param EIEIOType eieio_type: the type of message
        :param bool is_prefix: True if there is a prefix, False otherwise
        :param bool is_payload_base:
            True if there is a payload base, False otherwise
        :return: The size of the header in bytes
        :rtype: int
        """
        size = 2
        if is_prefix:
            size += 2
        if is_payload_base:
            size += eieio_type.key_bytes
        return size

    def increment_count(self) -> None:
        """
        Increase the count by 1.
        """
        self._count += 1

    def reset_count(self) -> None:
        """
        Resets the count back to zero.
        """
        self._count = 0

    @property
    def bytestring(self) -> bytes:
        """
        The byte-string of the header.

        :rtype: bytes
        """
        # Convert the flags to an int
        data = 0

        # the flag for prefix or not
        if self._prefix is not None:
            data |= 1 << 7

            # the prefix type
            data |= self._prefix_type.value << 6

        # the flag for payload prefix
        if self._payload_base is not None:
            data |= 1 << 5

        # the flag for time in payloads
        if self._is_time:
            data |= 1 << 4

        # The type of the packet
        data |= self._eieio_type.value << 2

        # The tag of the packet
        data |= self._tag

        # Convert the remaining data, depending on the various options
        if self._payload_base is None:
            if self._prefix is not None:
                return _PATTERN_BBH.pack(self._count, data, self._prefix)
            return _PATTERN_BB.pack(self._count, data)
        if (self._eieio_type == EIEIOType.KEY_PAYLOAD_16_BIT or
                self._eieio_type == EIEIOType.KEY_16_BIT):
            if self._prefix is not None:
                return _PATTERN_BBHH.pack(
                    self._count, data, self._prefix, self._payload_base)
            return _PATTERN_BBH.pack(self._count, data, self._payload_base)
        if (self._eieio_type == EIEIOType.KEY_PAYLOAD_32_BIT or
                self._eieio_type == EIEIOType.KEY_32_BIT):
            if self._prefix is not None:
                return _PATTERN_BBHI.pack(
                    self._count, data, self._prefix, self._payload_base)
            return _PATTERN_BBI.pack(self._count, data, self._payload_base)
        raise SpinnmanInvalidPacketException(
            "EIEIODataMessage", "unexpected EIEIO type code")

    @staticmethod
    def from_bytestring(data: bytes, offset: int) -> 'EIEIODataHeader':
        """
        Read an EIEIO data header from a byte-string.

        :param bytes data: The byte-string to be read
        :param int offset: The offset at which the data starts
        :return: an EIEIO header
        :rtype: EIEIODataHeader
        """
        (count, header_data) = _PATTERN_BB.unpack_from(data, offset)

        # Read the flags in the header
        prefix_flag = (header_data >> 7) & 1
        format_flag = (header_data >> 6) & 1
        payload_prefix_flag = (header_data >> 5) & 1
        payload_is_timestamp = (header_data >> 4) & 1
        message_type = (header_data >> 2) & 3
        tag = header_data & 3

        # Check for command header
        if prefix_flag == 0 and format_flag == 1:
            raise SpinnmanInvalidPacketException(
                "EIEIODataHeader",
                "The header indicates that this is a command header")

        # Convert the flags into types
        # pylint: disable=no-value-for-parameter
        eieio_type = EIEIOType(message_type)
        prefix_type = EIEIOPrefix(format_flag)

        prefix = None
        payload_prefix = None
        if payload_prefix_flag == 1:
            if (eieio_type == EIEIOType.KEY_16_BIT or
                    eieio_type == EIEIOType.KEY_PAYLOAD_16_BIT):
                if prefix_flag == 1:
                    (prefix, payload_prefix) = _PATTERN_HH.unpack_from(
                        data, offset + 2)
                else:
                    payload_prefix = _PATTERN_H.unpack_from(
                        data, offset + 2)[0]
            elif (eieio_type == EIEIOType.KEY_32_BIT or
                    eieio_type == EIEIOType.KEY_PAYLOAD_32_BIT):
                if prefix_flag == 1:
                    (prefix, payload_prefix) = _PATTERN_HI.unpack_from(
                        data, offset + 2)
                else:
                    payload_prefix = _PATTERN_I.unpack_from(
                        data, offset + 2)[0]
        else:
            if prefix_flag == 1:
                prefix = _PATTERN_H.unpack_from(data, offset + 2)[0]

        return EIEIODataHeader(
            eieio_type=eieio_type, tag=tag, prefix=prefix,
            prefix_type=prefix_type, payload_base=payload_prefix,
            is_time=bool(payload_is_timestamp), count=count)

    def __str__(self):
        return (f"EIEIODataHeader:prefix={self._prefix}:"
                f"prefix_type={self._prefix_type}:"
                f"payload_base={self._payload_base}:"
                "is_time={self._is_time}:type={self._eieio_type.value}:"
                "tag={self._tag}:count={self._count}")

    def __repr__(self):
        return self.__str__()



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
import unittest
from spinn_utilities.config_holder import set_config
from spinn_utilities.ping import Ping
from spinnman.constants import LOCAL_HOST


class BoardTestConfiguration(object):
    """
    Configuration to use for a test board
    """

    def __init__(self):
        self.remotehost = None
        self.auto_detect_bmp = None

    def set_up_remote_board(self, version: Optional[int] = None):
        """
        Gets a remote board to test, returning the first that it finds.

        Search order is
        - Local 4 Chip board
        - 48 Chip board "spinn-4.cs.man.ac.uk"
        - Local 48 Chip board if at 192.168.240.1
        - Virtual machine

        The first three ignore the version param the last needs it

        Sets the version field in the configs.

        :param version: Version for a virtual if no physical board found
        :type version: into or None
        :raises unittest.SkipTest:
            If no physical machine found and no version provided
        """
        if Ping.host_is_reachable("192.168.240.253"):
            self.remotehost = "192.168.240.253"
            set_config("Machine", "version", "3")
            self.auto_detect_bmp = False
        elif Ping.host_is_reachable("spinn-4.cs.man.ac.uk"):
            self.remotehost = "spinn-4.cs.man.ac.uk"
            set_config("Machine", "version", "5")
        elif Ping.host_is_reachable("192.168.240.1"):
            self.remotehost = "192.168.240.1"
            set_config("Machine", "version", "5")
        elif version is not None:
            self.remotehost = LOCAL_HOST
            set_config("Machine", "version", str(version))
        else:
            raise unittest.SkipTest("None of the test boards reachable")



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_machine.core_subsets import CoreSubsets
from spinnman.messages.scp.impl import ApplicationRun, CheckOKResponse
from .abstract_multi_connection_process import AbstractMultiConnectionProcess


class ApplicationRunProcess(AbstractMultiConnectionProcess[CheckOKResponse]):
    """
    A process to run an application.
    """
    __slots__ = ()

    def run(self, app_id: int, core_subsets: CoreSubsets, wait: bool):
        """
        Runs the application.

        :param int app_id:
        :param CoreSubsets core_subsets:
        :param bool wait:
`        """
        with self._collect_responses():
            for core_subset in core_subsets:
                self._send_request(ApplicationRun(
                    app_id, core_subset.x, core_subset.y,
                    core_subset.processor_ids, wait))



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class IOBuffer(object):
    """
    The contents of IOBUF for a core.
    """
    __slots__ = [
        "_iobuf",
        "_x", "_y", "_p"]

    def __init__(self, x: int, y: int, p: int, iobuf: str):
        """
        :param int x: The x-coordinate of a chip
        :param int y: The y-coordinate of a chip
        :param int p: The p-coordinate of a chip
        :param str iobuf: The contents of the buffer for the chip
        """
        self._x = x
        self._y = y
        self._p = p
        self._iobuf = iobuf

    @property
    def x(self) -> int:
        """
        The X-coordinate of the chip containing the core.

        :rtype: int
        """
        return self._x

    @property
    def y(self) -> int:
        """
        The Y-coordinate of the chip containing the core.

        :rtype: int
        """
        return self._y

    @property
    def p(self) -> int:
        """
        The ID of the core on the chip.

        :rtype: int
        """
        return self._p

    @property
    def iobuf(self) -> str:
        """
        The contents of the buffer.

        :rtype: str
        """
        return self._iobuf

    def __str__(self) -> str:
        value = ""
        for line in self._iobuf.split("\n"):
            value += f"{self._x}:{self._y}:{self._p:2n}: {line}\n"
        return value[:-1]



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Collection, Dict, Iterable, Optional, Set, cast
from spinn_utilities.ordered_set import OrderedSet
from spinn_machine import CoreSubsets, FrozenCoreSubsets
from spinnman.exceptions import SpinnmanInvalidParameterException
from .enums import ExecutableType


class ExecutableTargets(object):
    """
    Encapsulate the binaries and cores on which to execute them.
    """
    __slots__ = [
        "_all_core_subsets",
        "_targets",
        "_total_processors",
        "_binary_type_map"]

    __EMPTY_SUBSET = FrozenCoreSubsets()

    def __init__(self) -> None:
        self._targets: Dict[str, CoreSubsets] = dict()
        self._total_processors = 0
        self._all_core_subsets = CoreSubsets()
        self._binary_type_map: Dict[
            ExecutableType, Set[str]] = defaultdict(
                # Need to pretend!
                lambda: cast(Set, OrderedSet()))

    def add_subsets(
            self, binary: str, subsets: CoreSubsets,
            executable_type: Optional[ExecutableType] = None):
        """
        Add core subsets to a binary.

        :param str binary: the path to the binary needed to be executed
        :param ~spinn_machine.CoreSubsets subsets:
            the subset of cores that the binary needs to be loaded on
        :param ~spinnman.model.enum.ExecutableType executable_type:
            The type of this executable.
            ``None`` means don't record it.
        """
        try:
            for subset in subsets.core_subsets:
                for p in subset.processor_ids:
                    self.add_processor(binary, subset.x, subset.y, p)
        except AttributeError:
            if subsets is not None:
                raise
        if executable_type is not None:
            self._binary_type_map[executable_type].add(binary)

    def add_processor(
            self, binary: str, chip_x: int, chip_y: int, chip_p: int,
            executable_type: Optional[ExecutableType] = None):
        """
        Add a processor to the executable targets

        :param str binary: the binary path for executable
        :param int chip_x:
            the coordinate on the machine in terms of x for the chip
        :param int chip_y:
            the coordinate on the machine in terms of y for the chip
        :param int chip_p: the processor ID to place this executable on
        :param ~spinnman.model.enum.ExecutableType executable_type:
            the executable type for locating n cores of
        """
        if self.known(binary, chip_x, chip_y, chip_p):
            return
        if binary not in self._targets:
            self._targets[binary] = CoreSubsets()
        if executable_type is not None:
            self._binary_type_map[executable_type].add(binary)
        self._targets[binary].add_processor(chip_x, chip_y, chip_p)
        self._all_core_subsets.add_processor(chip_x, chip_y, chip_p)
        self._total_processors += 1

    def get_n_cores_for_executable_type(
            self, executable_type: ExecutableType) -> int:
        """
        Get the number of cores that the executable type is using.

        :param ~spinnman.model.enum.ExecutableType executable_type:
        :return: the number of cores using this executable type
        :rtype: int
        """
        return sum(
            len(self.get_cores_for_binary(aplx))
            for aplx in self._binary_type_map[executable_type])

    def get_binaries_of_executable_type(
            self, executable_type: ExecutableType) -> Iterable[str]:
        """
        Get the binaries of a given a executable type.

        :param ~spinnman.model.enum.ExecutableType executable_type:
            the executable type enum value
        :return: iterable of binaries with that executable type
        :rtype: iterable(str)
        """
        return self._binary_type_map[executable_type]

    def executable_types_in_binary_set(self) -> Iterable[ExecutableType]:
        """
        Get the executable types in the set of binaries.

        :return: iterable of the executable types in this binary set.
        :rtype:
            iterable(~spinnman.model.enum.ExecutableType)
        """
        return self._binary_type_map.keys()

    def get_cores_for_binary(self, binary: str) -> CoreSubsets:
        """
        Get the cores that a binary is to run on.

        :param str binary: The binary to find the cores for
        :rtype: ~spinn_machine.CoreSubsets
        """
        return self._targets.get(binary, self.__EMPTY_SUBSET)

    @property
    def binaries(self) -> Collection[str]:
        """
        The binaries of the executables.

        :rtype: iterable(str)
        """
        return self._targets.keys()

    @property
    def total_processors(self) -> int:
        """
        The total number of cores to be loaded.

        :rtype: int
        """
        return self._total_processors

    @property
    def all_core_subsets(self) -> CoreSubsets:
        """
        All the core subsets for all the binaries.

        :rtype: ~spinn_machine.CoreSubsets
        """
        return self._all_core_subsets

    def known(self, binary, chip_x, chip_y, chip_p) -> bool:
        """
        :param str binary:
        :param int chip_x:
        :param int chip_y:
        :param int chip_p:
        :rtype: bool
        """
        if not self._all_core_subsets.is_core(chip_x, chip_y, chip_p):
            return False
        # OK if and only if the chip is in this binary already
        if binary in self._targets:
            if self._targets[binary].is_core(chip_x, chip_y, chip_p):
                return True

        raise SpinnmanInvalidParameterException(
            f"x:{chip_x} y:{chip_y} p:{chip_p}", binary,
            "Already associated with a different binary")



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from threading import RLock
import time
from types import TracebackType
from typing import Callable, Dict, Generic, List, Optional, TypeVar, cast
from typing_extensions import TypeAlias
from spinnman.messages.scp.enums import SCPResult
from spinnman.exceptions import SpinnmanTimeoutException, SpinnmanIOException
from spinnman.constants import SCP_TIMEOUT, N_RETRIES
from spinnman.connections.udp_packet_connections import SCAMPConnection
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.abstract_messages import AbstractSCPResponse

#: Type of responses.
#: :meta private:
R = TypeVar("R", bound=AbstractSCPResponse)
#: Type of response-accepting callbacks.
#: :meta private:
CB: TypeAlias = Callable[[R], None]
#: Type of error-handling callbacks.
#: :meta private:
ECB: TypeAlias = Callable[
    [AbstractSCPRequest[R], Exception, TracebackType, SCAMPConnection], None]

MAX_SEQUENCE = 65536
RETRY_CODES = frozenset([
    SCPResult.RC_TIMEOUT, SCPResult.RC_P2P_TIMEOUT, SCPResult.RC_LEN,
    SCPResult.RC_P2P_NOREPLY, SCPResult.RC_P2P_BUSY])

# Keep a global track of the sequence numbers used
_next_sequence = 0
_next_sequence_lock = RLock()


class SCPRequestPipeLine(Generic[R]):
    """
    Allows a set of SCP requests to be grouped together in a communication
    across a number of channels for a given connection.

    This class implements an SCP windowing, first suggested by Andrew Mundy.
    This extends the idea by having both send and receive windows.
    These are represented by the n_channels and the
    intermediate_channel_waits parameters respectively.  This seems to
    help with the timeout issue; when a timeout is received, all requests
    for which a reply has not been received can also timeout.
    """
    __slots__ = (
        "_callbacks",
        "_connection",
        "_error_callbacks",
        "_in_progress",
        "_intermediate_channel_waits",
        "_n_channels",
        "_n_resent",
        "_n_retries",
        "_n_retry_code_resent",
        "_n_timeouts",
        "_non_fail_retry_codes",
        "_packet_timeout",
        "_retry_reason",
        "_request_data",
        "_requests",
        "_retries",
        "_send_time")

    def __init__(self, connection: SCAMPConnection, n_channels=1,
                 intermediate_channel_waits=0,
                 n_retries=N_RETRIES, packet_timeout=SCP_TIMEOUT,
                 non_fail_retry_codes=None):
        """
        :param SCAMPConnection connection:
            The connection over which the communication is to take place
        :param int n_channels: The number of requests to send before checking
            for responses.  If `None`, this will be determined automatically
        :param int intermediate_channel_waits: The number of outstanding
            responses to wait for before continuing sending requests.
            If `None`, this will be determined automatically
        :param int n_retries: The number of times to resend any packet for any
            reason before an error is triggered
        :param float packet_timeout: The number of elapsed seconds after
            sending a packet before it is considered a timeout.
        :param non_fail_retry_codes: Codes that could retry but won't fail, or
            None if there are no such codes
        :type non_fail_retry_codes: set(SCPResult) or None
        """
        self._connection = connection
        self._n_channels = n_channels
        self._intermediate_channel_waits = intermediate_channel_waits
        self._n_retries = n_retries
        self._packet_timeout = packet_timeout

        if (self._n_channels is not None and
                self._intermediate_channel_waits is None):
            self._intermediate_channel_waits = self._n_channels - 8
            if self._intermediate_channel_waits < 0:
                self._intermediate_channel_waits = 0

        # A dictionary of sequence number -> requests in progress
        self._requests: Dict[int, AbstractSCPRequest] = dict()
        self._request_data: Dict[int, bytes] = dict()

        # A dictionary of sequence number -> number of retries for the packet
        self._retries: Dict[int, int] = dict()

        # A dictionary of sequence number -> callback function for response
        self._callbacks: Dict[int, Optional[CB]] = dict()

        # A dictionary of sequence number -> callback function for errors
        self._error_callbacks: Dict[int, ECB] = dict()

        # A dictionary of sequence number -> retry reason
        self._retry_reason: Dict[int, List[str]] = dict()

        # The number of responses outstanding
        self._in_progress = 0

        # The number of timeouts that occurred
        self._n_timeouts = 0

        # The number of packets that have been resent
        self._n_resent = 0
        self._n_retry_code_resent = 0
        self._non_fail_retry_codes = non_fail_retry_codes
        if self._non_fail_retry_codes is None:
            self._non_fail_retry_codes = set()

        # self._token_bucket = TokenBucket(43750, 4375000)
        # self._token_bucket = TokenBucket(3408, 700000)

    @staticmethod
    def __get_next_sequence_number() -> int:
        """
        Get the next number from the global sequence, applying appropriate
        wrapping rules as the sequence numbers have a fixed number of bits.

        :return: The next number in the sequence.
        :rtype: int
        """
        # pylint: disable=global-statement
        global _next_sequence
        with _next_sequence_lock:
            sequence = _next_sequence
            _next_sequence = (sequence + 1) % MAX_SEQUENCE
        return sequence

    def send_request(
            self, request: AbstractSCPRequest[R], callback: Optional[CB],
            error_callback: ECB):
        """
        Add an SCP request to the set to be sent.

        :param AbstractSCPRequest request: The SCP request to be sent
        :param ~collections.abc.Callable callback:
            A callback function to call when the response has been received;
            takes a :py:class:`SCPResponse` as a parameter, or `None` if the
            response doesn't need to be processed
        :param ~collections.abc.Callable error_callback:
            A callback function to call when an error is found when processing
            the message; takes the original :py:class:`AbstractSCPRequest`, the
            exception caught and a list of tuples of (filename, line number,
            function name, text) as a traceback
        """
        # If the connection has not been measured
        if self._n_channels is None:
            if self._connection.is_ready_to_receive():
                self._n_channels = self._in_progress + 8
                if self._n_channels < 12:
                    self._n_channels = 12
                self._intermediate_channel_waits = self._n_channels - 8

        # If all the channels are used, start to receive packets
        while (self._n_channels is not None and
                self._in_progress >= self._n_channels):
            self._do_retrieve(
                self._intermediate_channel_waits, self._packet_timeout)

        # Get the next sequence to be used
        sequence = self.__get_next_sequence_number()

        # Update the packet and store required details
        request.scp_request_header.sequence = sequence
        request_data = self._connection.get_scp_data(request)
        self._requests[sequence] = request
        self._request_data[sequence] = request_data
        self._retries[sequence] = self._n_retries
        self._callbacks[sequence] = callback
        self._error_callbacks[sequence] = error_callback
        self._retry_reason[sequence] = list()

        # Send the request, keeping track of how many are sent
        # self._token_bucket.consume(284)
        self._connection.send(request_data)
        self._in_progress += 1

    def finish(self) -> None:
        """
        Indicate the end of the packets to be sent.  This must be called
        to ensure that all responses are received and handled.
        """
        while self._in_progress > 0:
            self._do_retrieve(0, self._packet_timeout)

    @property
    def n_timeouts(self) -> int:
        """
        The number of timeouts that occurred.

        :rtype: int
        """
        return self._n_timeouts

    @property
    def n_channels(self) -> int:
        """
        The number of requests to send before checking for responses.

        :rtype: int
        """
        return self._n_channels

    @property
    def n_resent(self) -> int:
        """
        The number of packets that have been resent.

        :rtype: int
        """
        return self._n_resent

    @property
    def n_retry_code_resent(self) -> int:
        """
        The number of resends due to reasons for which automated retry is
        the correct response in-protocol.

        :rtype: int
        """
        return self._n_retry_code_resent

    def _remove_record(self, seq: int) -> None:
        if seq in self._requests:
            del self._requests[seq]
        del self._request_data[seq]
        del self._retries[seq]
        del self._callbacks[seq]
        del self._error_callbacks[seq]
        del self._retry_reason[seq]

    def _single_retrieve(self, timeout: float):
        # Receive the next response
        result, seq, raw_data, offset = \
            self._connection.receive_scp_response(timeout)

        # Only process responses which have matching requests
        if seq in self._requests:
            self._in_progress -= 1
            request_sent = self._requests[seq]

            # If the response can be retried, retry it
            if result in RETRY_CODES:
                try:
                    time.sleep(0.1)
                    self._resend(seq, request_sent, str(result))
                    self._n_retry_code_resent += 1
                    return
                except Exception as e:  # pylint: disable=broad-except
                    if result not in self._non_fail_retry_codes:
                        self._error_callbacks[seq](
                            request_sent, e,
                            cast(TracebackType, sys.exc_info()[2]),
                            self._connection)
                        self._remove_record(seq)
                        return

            # No retry is possible and not failed - try constructing the result
            try:
                response = request_sent.get_scp_response()
                response.read_bytestring(raw_data, offset)
                cb = self._callbacks[seq]
                if cb is not None:
                    cb(response)
            except Exception as e:  # pylint: disable=broad-except
                self._error_callbacks[seq](
                    request_sent, e,
                    cast(TracebackType, sys.exc_info()[2]),
                    self._connection)

            # Remove the sequence from the outstanding responses
            self._remove_record(seq)

    def _handle_receive_timeout(self) -> None:
        self._n_timeouts += 1

        # If there is a timeout, all packets remaining are resent
        to_remove = list()
        for seq, request_sent in self._requests.items():
            self._in_progress -= 1
            try:
                self._resend(seq, request_sent, "timeout")
            except Exception as e:  # pylint: disable=broad-except
                self._error_callbacks[seq](
                    request_sent, e, cast(TracebackType, sys.exc_info()[2]),
                    self._connection)
                to_remove.append(seq)

        for seq in to_remove:
            self._remove_record(seq)

    def _resend(self, seq: int, request_sent, reason: str):
        if self._retries[seq] <= 0:
            # Report timeouts as timeout exception

            self._retry_reason[seq].append(reason)
            if all(reason == "timeout" for reason in self._retry_reason[seq]):
                raise SpinnmanTimeoutException(
                    request_sent,
                    self._packet_timeout)

            # Report any other exception
            raise SpinnmanIOException(
                f"Errors sending request {request_sent} to "
                f"{request_sent.sdp_header.destination_chip_x}, "
                f"{request_sent.sdp_header.destination_chip_y}, "
                f"{request_sent.sdp_header.destination_cpu} over "
                f"{self._n_retries} retries: {self._retry_reason[seq]}")

        # If the request can be retried, retry it
        self._retries[seq] -= 1
        self._in_progress += 1
        self._requests[seq] = request_sent
        self._retry_reason[seq].append(reason)
        self._connection.send(self._request_data[seq])
        self._n_resent += 1

    def _do_retrieve(self, n_packets: int, timeout: float):
        """
        Receives responses until there are only n_packets responses left.

        :param int n_packets:
            The number of packets that can remain after running
        """
        # While there are still more packets in progress than some threshold
        while self._in_progress > n_packets:
            try:
                # Receive the next response
                self._single_retrieve(timeout)
            except SpinnmanTimeoutException:
                self._handle_receive_timeout()



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from typing import BinaryIO, Callable
import numpy
from numpy import uint8, uint32
from spinn_utilities.typing.coords import XYP
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.impl import WriteLink, WriteMemory
from spinnman.messages.scp.impl import CheckOKResponse
from spinnman.constants import UDP_MESSAGE_MAX_SIZE
from .abstract_multi_connection_process import AbstractMultiConnectionProcess

_UNSIGNED_WORD = 0xFFFFFFFF


class WriteMemoryProcess(AbstractMultiConnectionProcess[CheckOKResponse]):
    """
    A process for writing memory on a SpiNNaker chip.
    """
    __slots__ = ()
    # pylint: disable=too-many-arguments

    def write_memory_from_bytearray(
            self, coordinates: XYP, base_address: int, data: bytes,
            offset: int, n_bytes: int, get_sum: bool = False) -> int:
        """
        Writes memory onto a SpiNNaker chip from a bytearray.

        :param tuple(int,int,int) coordinates:
            The X,Y,P coordinates of the core that will write to memory.
        :param int base_address: the address in SDRAM to start writing
        :param data: the data to write
        :type data: bytearray or bytes
        :param int offset: where in the data to start writing from
        :param int n_bytes: how much data to write
        :param bool get_sum: whether to return a checksum or 0
        :return: the data checksum or 0 if get_sum is False
        :rtype: int
        """
        return self._write_memory_from_bytearray(
            base_address, data, offset, n_bytes,
            functools.partial(WriteMemory, coordinates), get_sum)

    def write_link_memory_from_bytearray(
            self, coordinates: XYP, link: int, base_address: int, data: bytes,
            offset: int, n_bytes: int, get_sum: bool = False) -> int:
        """
        Writes memory onto a neighbour of a SpiNNaker chip from a bytearray.

        :param tuple(int,int,int) coordinates:
            The X,Y,P coordinates of the core that will write to its
            neighbour's memory.
        :param int link:
            Along which link is the neighbour.
        :param int base_address: the address in SDRAM to start writing
        :param data: the data to write
        :type data: bytearray or bytes
        :param int offset: where in the data to start writing from
        :param int n_bytes: how much data to write
        :param bool get_sum: whether to return a checksum or 0
        :return: the data checksum or 0 if get_sum is False
        :rtype: int
        """
        return self._write_memory_from_bytearray(
            base_address, data, offset, n_bytes,
            functools.partial(WriteLink, coordinates, link), get_sum)

    def write_memory_from_reader(
            self, coordinates: XYP, base_address: int, reader: BinaryIO,
            n_bytes: int, get_sum: bool = False) -> int:
        """
        Writes memory onto a SpiNNaker chip from a reader.

        :param tuple(int,int,int) coordinates:
            The X,Y,P coordinates of the core that will write to memory.
        :param int base_address: the address in SDRAM to start writing
        :param reader: the readable object containing the data to write
        :type reader: ~io.RawIOBase or ~io.BufferedIOBase
        :param int n_bytes: how much data to write
        :param bool get_sum: whether to return a checksum or 0
        :return: the data checksum or 0 if get_sum is False
        :rtype: int
        """
        return self._write_memory_from_reader(
            base_address, reader, n_bytes,
            functools.partial(WriteMemory, coordinates), get_sum)

    def write_link_memory_from_reader(
            self, coordinates: XYP, link: int, base_address: int,
            reader: BinaryIO, n_bytes: int, get_sum: bool = False) -> int:
        """
        Writes memory onto a neighbour of a SpiNNaker chip from a reader.

        :param tuple(int,int,int) coordinates:
            The X,Y,P coordinates of the core that will write to its
            neighbour's memory. The P coordinate is normally 0; no reason to
            not use SCAMP for this.
        :param int link:
            Along which link is the neighbour.
        :param int base_address: the address in SDRAM to start writing
        :param reader: the readable object containing the data to write
        :type reader: ~io.RawIOBase or ~io.BufferedIOBase
        :param int n_bytes: how much data to write
        :param bool get_sum: whether to return a checksum or 0
        :return: the data checksum or 0 if get_sum is False
        :rtype: int
        """
        return self._write_memory_from_reader(
            base_address, reader, n_bytes,
            functools.partial(WriteLink, coordinates, link), get_sum)

    def _write_memory_from_bytearray(
            self, base_address: int, data: bytes, data_offset: int,
            n_bytes: int, packet_class: Callable[
                [int, bytes], AbstractSCPRequest[CheckOKResponse]],
            get_sum: bool) -> int:
        offset = 0
        n_bytes_to_write = int(n_bytes)
        with self._collect_responses():
            while n_bytes_to_write > 0:
                bytes_to_send = min(n_bytes_to_write, UDP_MESSAGE_MAX_SIZE)
                data_array = data[data_offset:data_offset + bytes_to_send]

                self._send_request(
                    packet_class(base_address + offset, data_array))

                n_bytes_to_write -= bytes_to_send
                offset += bytes_to_send
                data_offset += bytes_to_send
        if not get_sum:
            return 0
        np_data = numpy.array(data, dtype=uint8)
        np_sum = int(numpy.sum(np_data.view(uint32), dtype=uint32))
        return np_sum & _UNSIGNED_WORD

    def _write_memory_from_reader(
            self, base_address: int, reader: BinaryIO, n_bytes: int,
            packet_class: Callable[
                [int, bytes], AbstractSCPRequest[CheckOKResponse]],
            with_sum: bool) -> int:
        offset = 0
        n_bytes_to_write = int(n_bytes)
        chksum = 0
        with self._collect_responses():
            while n_bytes_to_write > 0:
                data_array = reader.read(
                    min(n_bytes_to_write, UDP_MESSAGE_MAX_SIZE))
                bytes_to_send = len(data_array)
                self._send_request(packet_class(
                    base_address + offset, data_array))

                n_bytes_to_write -= bytes_to_send
                offset += bytes_to_send

                if with_sum:
                    np_data = numpy.frombuffer(data_array, dtype=uint8)
                    np_sum = int(numpy.sum(np_data.view(uint32)))
                    chksum = (chksum + np_sum) & _UNSIGNED_WORD

        return chksum
